import unittest

import healpy as hp
import numpy as np

import rubin_scheduler.scheduler.basis_functions as bf
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.surveys import BlobSurvey
from rubin_scheduler.scheduler.utils import Footprint


class TestComCam(unittest.TestCase):
    def test_comcam(self):
        """Make a simple r and g survey to run with ComCam."""
        # Specify the HEALpix resolution we want to do everything in.
        # Need to turn up for comcam
        nside = 512
        mo = ModelObservatory(nside=nside, downtimes="ideal", cloud_data="ideal")
        conditions = mo.return_conditions()

        # Let's start out with a simple 2-band footprint where
        # we want 1 observations in r for every 2 observations in g
        blank_map = np.zeros(hp.nside2npix(nside)) + np.nan
        # cut down to a very narrow band
        indx = np.where((conditions.dec < np.radians(-15)) & (conditions.dec > np.radians(-20)))
        simple_fp = {"r": blank_map + 0, "g": blank_map + 0}
        simple_fp["r"][indx] = 1
        simple_fp["g"][indx] = 1

        fp = Footprint(mo.mjd_start, mo.sun_ra_start, nside=nside)
        for bandname in simple_fp:
            fp.set_footprint(bandname, simple_fp[bandname])

        footprint_weight = 1.0
        m5_weight = 0.0

        detailers = []

        red_fp_basis = bf.FootprintBasisFunction(bandname="r", footprint=fp, nside=nside)
        m5_basis_r = bf.M5DiffBasisFunction(bandname="r", nside=nside)
        red_survey = BlobSurvey(
            [red_fp_basis, m5_basis_r],
            [footprint_weight, m5_weight],
            bandname1="r",
            survey_name="r_blob",
            nside=nside,
            camera="comcam",
            grow_blob=False,
            detailers=detailers,
            dither=False,
            twilight_scale=False,
        )

        blue_fp_basis = bf.FootprintBasisFunction(bandname="g", footprint=fp, nside=nside)
        m5_basis_g = bf.M5DiffBasisFunction(bandname="g", nside=nside)
        blue_survey = BlobSurvey(
            [blue_fp_basis, m5_basis_g],
            [footprint_weight, m5_weight],
            bandname1="g",
            survey_name="g_blob",
            nside=nside,
            camera="comcam",
            grow_blob=False,
            detailers=detailers,
            dither=False,
            twilight_scale=False,
        )

        scheduler = CoreScheduler([red_survey, blue_survey], nside=nside, camera="comcam")

        mo, scheduler, observations = sim_runner(mo, scheduler, sim_duration=0.5, verbose=True)

        assert len(observations) > 100
        assert np.size(np.where(observations["band"] == "r")[0]) > 0
        assert np.size(np.where(observations["band"] == "g")[0]) > 0


if __name__ == "__main__":
    unittest.main()



import numpy as np
from astropy.coordinates import EarthLocation, get_body, solar_system_ephemeris
from astropy.time import Time

from rubin_scheduler.utils import Site

if __name__ == "__main__":
    mjd_start = 59853.5 - 3.0 * 365.25
    duration = 30.0 * 365.25
    pad_around = 40
    t_step = 1.0 / 24.0  # Start with 1-hour timesteps.

    mjds = np.arange(mjd_start - pad_around, duration + mjd_start + pad_around + t_step, t_step)

    planet_names = ["venus", "mars", "jupiter", "saturn"]

    names = ["mjd"]
    for pn in planet_names:
        names.append(pn + "_RA")
        names.append(pn + "_dec")

    types = [float] * len(names)

    planet_loc = np.zeros(mjds.size, dtype=list(zip(names, types)))
    planet_loc["mjd"] = np.arange(mjd_start - pad_around, duration + mjd_start + pad_around + t_step, t_step)
    site = Site("LSST")
    location = EarthLocation(lat=site.latitude, lon=site.longitude, height=site.height)
    t_sparse = Time(mjds, format="mjd", location=location)

    # Should get dowloaded and chached if needed.
    # https://docs.astropy.org/en/stable/coordinates/solarsystem.html
    # might need to `pip install jplephem`
    solar_system_ephemeris.set("de432s")

    for pn in planet_names:
        print("computing locations of %s" % pn)
        body = get_body(pn, t_sparse, location)
        body_icrs = body.icrs
        planet_loc[pn + "_RA"] = body_icrs.ra.rad
        planet_loc[pn + "_dec"] = body_icrs.dec.rad

    # I could crop off the times when the sun is up, but meh, bits are cheap
    np.savez("planet_locations.npz", planet_loc=planet_loc)



__all__ = ("DescDdf", "generate_desc_dd_surveys")

import numpy as np

import rubin_scheduler.scheduler.basis_functions as basis_functions
from rubin_scheduler.scheduler.surveys import BaseSurvey
from rubin_scheduler.scheduler.utils import ObservationArray


class DescDdf(BaseSurvey):
    """DDF survey based on Scolnic et al Cadence White Paper."""

    def __init__(
        self,
        basis_functions,
        RA,
        dec,
        sequences=None,
        exptime=30.0,
        nexp=1,
        ignore_obs=None,
        survey_name="DD_DESC",
        reward_value=101.0,
        readtime=2.0,
        band_change_time=120.0,
        nside=None,
        flush_pad=30.0,
        seed=42,
        detailers=None,
    ):
        super(DescDdf, self).__init__(
            nside=nside,
            basis_functions=basis_functions,
            detailers=detailers,
            ignore_obs=ignore_obs,
        )

        self.ra = np.radians(RA)
        self.ra_hours = RA / 360.0 * 24.0
        self.dec = np.radians(dec)
        self.survey_name = survey_name
        self.reward_value = reward_value
        self.flush_pad = flush_pad / 60.0 / 24.0  # To days

        self.simple_obs = ObservationArray()
        self.simple_obs["RA"] = np.radians(RA)
        self.simple_obs["dec"] = np.radians(dec)
        self.simple_obs["exptime"] = exptime
        self.simple_obs["nexp"] = nexp
        self.simple_obs["scheduler_note"] = survey_name

        # Define the sequences we would like to do
        if sequences is None:
            self.sequences = [{"u": 2, "g": 2, "r": 4, "i": 8}, {"z": 25, "y": 4}, None]
        else:
            self.sequences = sequences

        self.approx_times = []
        for sequence in self.sequences:
            if sequence is None:
                self.approx_times.append(0)
            else:
                n_exp_in_seq = np.sum(list(sequence.values()))
                time_needed = band_change_time * len(sequence.keys())
                time_needed += exptime * n_exp_in_seq
                time_needed += readtime * n_exp_in_seq * nexp
                self.approx_times.append(time_needed / 3600.0 / 24.0)

        # Track what we last tried to do
        # XXX-this should probably go into self.extra_features or
        # something for consistency.
        self.sequence_index = 0
        self.last_night_observed = -100

    def check_continue(self, observation, conditions):
        # feasibility basis functions?
        """
        This method enables external calls to check if a given
        observations that belongs to this survey is
        feasible or not. This is called once a sequence has started to
        make sure it can continue.

        XXX--TODO:  Need to decide if we want to develope check_continue
         or instead hold the sequence in the survey, and be able to check
         it that way.
        """

        result = True
        return result

    def _check_feasibility(self, conditions):
        """
        Check if the survey is feasable in the current conditions
        """

        # No more if we've already observed this night
        if self.last_night_observed == conditions.night:
            return False

        # Advance the sequence index if we have skipped a day intentionally
        if (self.sequences[self.sequence_index] is None) & (conditions.night - self.last_night_observed > 1):
            self.sequence_index = (self.sequence_index + 1) % len(self.sequences)

        # If we want to skip this day
        if self.sequences[self.sequence_index] is None:
            return False

        # The usual basis function checks
        for bf in self.basis_functions:
            result = bf.check_feasibility(conditions)
            if not result:
                return result
        return result

    def calc_reward_function(self, conditions):
        result = -np.inf
        if self._check_feasibility(conditions):
            result = self.reward_value
        return result

    def generate_observations_rough(self, conditions):
        result = []
        for key in self.sequences[self.sequence_index]:
            # Just skip adding the z-band ones if it's not loaded
            if key in conditions.mounted_bands:
                for i in np.arange(self.sequences[self.sequence_index][key]):
                    temp_obs = self.simple_obs.copy()
                    temp_obs["band"] = key
                    # XXX--need to set flush by mjd
                    result.append(temp_obs)
        for i, obs in enumerate(result):
            result[i]["flush_by_mjd"] = (
                conditions.mjd + self.approx_times[self.sequence_index] + self.flush_pad
            )
        # Just assuming this sequence gets observed.
        self.last_night_observed = conditions.night
        self.sequence_index = (self.sequence_index + 1) % len(self.sequences)
        return result


def desc_dd_bfs(RA, dec, survey_name, ha_limits, frac_total=0.0185):
    """
    Convienence function to generate all the feasibility basis functions
    """
    bfs = []
    bfs.append(basis_functions.Not_twilight_basis_function(sun_alt_limit=-18))
    bfs.append(basis_functions.Time_to_twilight_basis_function(time_needed=30.0))
    bfs.append(basis_functions.Hour_Angle_limit_basis_function(RA=RA, ha_limits=ha_limits))
    bfs.append(basis_functions.Rising_more_basis_function(RA=RA))
    bfs.append(basis_functions.Clouded_out_basis_function())

    return bfs


def generate_desc_dd_surveys(nside=None, nexp=1, detailers=None):
    surveys = []

    # ELAIS S1
    RA = 9.45
    dec = -44.0
    survey_name = "DD:ELAISS1"
    ha_limits = ([0.0, 1.18], [21.82, 24.0])
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)
    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # XMM-LSS
    survey_name = "DD:XMM-LSS"
    RA = 35.708333
    dec = -4 - 45 / 60.0
    ha_limits = ([0.0, 1.3], [21.7, 24.0])
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)

    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )
    # Extended Chandra Deep Field South
    RA = 53.125
    dec = -28.0 - 6 / 60.0
    survey_name = "DD:ECDFS"
    ha_limits = [[0.5, 3.0], [20.0, 22.5]]
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)
    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # COSMOS
    RA = 150.1
    dec = 2.0 + 10.0 / 60.0 + 55 / 3600.0
    survey_name = "DD:COSMOS"
    ha_limits = ([0.0, 1.5], [21.5, 24.0])
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)
    # have a special sequence for COSMOS
    sequences = [{"g": 2, "r": 4, "i": 8}, {"z": 25, "y": 4}]
    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
            sequences=sequences,
        )
    )

    # Just do the two Euclid fields independently for now
    survey_name = "DD:EDFSa"
    RA = 58.97
    dec = -49.28
    ha_limits = ([0.0, 1.5], [23.0, 24.0])
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)
    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    survey_name = "DD:EDFSb"
    RA = 63.6
    dec = -47.60
    ha_limits = ([0.0, 1.5], [23.0, 24.0])
    bfs = desc_dd_bfs(RA, dec, survey_name, ha_limits)
    surveys.append(
        DescDdf(
            bfs,
            RA,
            dec,
            survey_name=survey_name,
            reward_value=100,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    return surveys

    pass



import copy
import unittest

import numpy as np

import rubin_scheduler.scheduler.detailers as detailers
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import CurrentAreaMap, ObservationArray, TargetoO
from rubin_scheduler.utils import DEFAULT_NSIDE, _ra_dec2_hpid


class TestDetailers(unittest.TestCase):
    def test_basics(self):
        """Test basic detailer functionality"""

        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        dec = np.radians(-20)
        ra = np.arange(0, 2 * np.pi, np.pi / 4)

        obs_array = ObservationArray(n=ra.size)
        obs_array["band"] = "r"
        obs_array["RA"] = ra
        obs_array["dec"] = dec
        obs_array["mjd"] = 59000.0
        obs_array["exptime"] = 30.0
        obs_array["scheduler_note"] = "test_note, a"

        hpid = _ra_dec2_hpid(DEFAULT_NSIDE, ra, dec)

        indx = np.isfinite(conditions.m5_depth["r"][hpid])
        obs_array = obs_array[indx]

        det_list = [
            detailers.VaryExptDetailer,
            detailers.ZeroRotDetailer,
            detailers.Comcam90rotDetailer,
            detailers.Rottep2RotspDesiredDetailer,
            detailers.CloseAltDetailer,
            detailers.TakeAsPairsDetailer,
            detailers.TwilightTripleDetailer,
            detailers.FlushForSchedDetailer,
            detailers.BandNexp,
            detailers.FixedSkyAngleDetailer,
            detailers.ParallacticRotationDetailer,
            detailers.FlushByDetailer,
            detailers.RandomBandDetailer,
            detailers.TrackingInfoDetailer,
            detailers.AltAz2RaDecDetailer,
            detailers.DitherDetailer,
            detailers.CameraRotDetailer,
            detailers.CameraSmallRotPerObservationListDetailer,
            detailers.BandToFilterDetailer,
            detailers.TagRadialDetailer,
            detailers.LabelDDFDetailer,
            detailers.TruncatePreTwiDetailer,
        ]

        for det in det_list:
            obs = copy.deepcopy(obs_array)
            live_det = det()
            result = live_det(obs, conditions)
            assert len(result) > 0

        # test the Euclid detailer
        live_det = detailers.EuclidDitherDetailer()
        # No "EDFS" should raise an error
        with self.assertRaises(ValueError):
            obs = copy.deepcopy(obs_array)
            live_det(obs, conditions)

        # Update so detailer should run
        obs = copy.deepcopy(obs_array)
        obs["scheduler_note"] = "DD:EDFS_a, 1212, 12, abc"
        result = live_det(obs, conditions)
        assert len(result) > 0

    def test_start_field(self):
        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        scheduler_note = "prepended"

        obs_to_prepend = ObservationArray(n=3)
        obs_to_prepend["RA"] = np.radians(20)
        obs_to_prepend["band"] = "r"
        obs_to_prepend["scheduler_note"] = scheduler_note

        obs_reg = ObservationArray(n=1)
        det = detailers.StartFieldSequenceDetailer(obs_to_prepend, scheduler_note=scheduler_note)
        obs_out = det(obs_reg, conditions)

        assert len(obs_out) == 4

    def test_delta_dither(self):
        obs = ObservationArray(2)
        obs["RA"] = np.radians(0)
        obs["dec"] = np.radians(20)

        det = detailers.DeltaCoordDitherDetailer(np.array([0.0, 0.0]), np.array([-10.0, 10.0]))

        output = det(obs, [])

        assert np.size(output) == 4
        assert np.unique(output["RA"]) == np.unique(obs["RA"])

        obs["RA"] = np.radians(90)
        det = detailers.DeltaCoordDitherDetailer(np.array([0.0, 0.0]), np.array([-10.0, 10.0]))
        output = det(obs, [])
        assert np.size(output) == 4
        assert np.unique(output["RA"]) == np.unique(obs["RA"])

        ra_step = 5.0
        det = detailers.DeltaCoordDitherDetailer(np.array([-ra_step, ra_step]), np.array([0.0, 0.0]))
        output = det(obs, [])

        # Make sure ra diff is larger
        assert np.all(np.abs(output["RA"] - obs["RA"][0]) > np.radians(ra_step))

        # Try having rotation as well
        det = detailers.DeltaCoordDitherDetailer(
            np.array([-ra_step, ra_step]), np.array([0.0, 0.0]), delta_rotskypos=np.array([5.0, 10.0])
        )
        output = det(obs, [])

        assert np.size(output) == 4
        assert np.size(np.unique(output["rotSkyPos_desired"])) == 2

        # Make sure one-element works
        obs = ObservationArray(1)
        obs["RA"] = np.radians(0)
        obs["dec"] = np.radians(20)
        det = detailers.DeltaCoordDitherDetailer(np.array([0.0, 0.0]), np.array([-10.0, 10.0]))
        output = det(obs, [])

        assert np.size(output) == 2
        assert np.unique(output["RA"]) == np.unique(obs["RA"])

        # Test going to the pole
        obs = ObservationArray(2)
        obs["RA"] = np.radians(0)
        obs["dec"] = np.radians(-90)

        det = detailers.DeltaCoordDitherDetailer(np.array([-1.0, 1.0, -1, 1]), np.array([1.0, 1.0, -1, -1]))
        output = det(obs, [])

        # This should make a grid all at the same dec
        assert np.size(np.unique(output["dec"])) == 1
        assert output["dec"][0] > obs["dec"][0]

    def test_random_band(self):
        obs = ObservationArray(1)
        obs["band"] = "r"

        det = detailers.RandomBandDetailer(bands="iz")

        conditions = Conditions(mjd=3, survey_start_mjd=0)
        conditions.mounted_bands = ["i", "z"]

        out_obs = det(obs, conditions)
        assert (out_obs["band"] == "i") | (out_obs["band"] == "z")

        # Check that we fall back properly
        conditions.mounted_bands = ["r", "g", "u", "y"]
        det = detailers.RandomBandDetailer(bands="iz", fallback_order="y")
        out_obs = det(obs, conditions)

        assert out_obs["band"] == "y"

    def test_band_sort_detailer(self):
        obs = ObservationArray(5)
        input_bands = ["r", "g", "r", "g", "z"]
        for i in range(len(obs)):
            obs[i]["band"] = input_bands[i]

        conditions = Conditions()
        conditions.mounted_bands = ["g", "r", "i", "z", "y"]
        conditions.current_band = "g"

        det = detailers.BandSortDetailer(desired_band_order="rgz", loaded_first=True)
        out_obs = det(obs, conditions)
        self.assertTrue(np.all(out_obs["band"] == ["g", "g", "r", "r", "z"]))

        det = detailers.BandSortDetailer(desired_band_order="rgz", loaded_first=False)
        out_obs = det(obs, conditions)
        self.assertTrue(np.all(out_obs["band"] == ["r", "r", "g", "g", "z"]))

    def test_bandtofilter(self):
        obs = ObservationArray(3)
        obs["band"][0] = "r"
        obs["band"][1] = "g"
        obs["band"][2] = "g"

        detailer = detailers.BandToFilterDetailer({})
        output = detailer(obs, [])
        assert np.all(output["band"] == output["filter"])

        filtername_dict = {"r": "r_03", "g": "g_01"}
        detailer = detailers.BandToFilterDetailer(filtername_dict)
        output = detailer(obs, [])

        for out in output:
            assert out["filter"] == filtername_dict[out["band"]]

    def test_copycol(self):
        obs = ObservationArray(3)
        obs["band"][0] = "r"
        obs["band"][1] = "g"
        obs["band"][2] = "g"

        detailer = detailers.CopyValueDetailer("band", "filter")

        output = detailer(obs, None)

        assert np.array_equal(output["band"], output["filter"])

    def test_event(self):
        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        detailer = detailers.BandPickToODetailer(
            band_start="g", band_end="r", distance_limit=10.0, check_mounted=True
        )

        obs = ObservationArray(3)
        obs["RA"] = np.radians(0.0)
        obs["dec"] = np.radians(-20.0)
        obs["band"] = "g"

        too_event = TargetoO(
            100, None, 10, 10, ra_rad_center=None, dec_rad_center=None, too_type=None, posterior_distance=1e9
        )

        # Should not do anything if distance to large
        out_obs = detailer(obs, conditions, target_o_o=too_event)
        assert np.all(out_obs["band"] == "g")

        # Now it is small enough
        too_event.posterior_distance = 0
        out_obs = detailer(obs, conditions, target_o_o=too_event)
        assert np.all(out_obs["band"] == "r")

        # reset
        obs["band"] = "g"

        # does nothing if not mounted
        conditions.mounted_bands = ["a", "b", "c"]
        out_obs = detailer(obs, conditions, target_o_o=too_event)
        assert np.all(out_obs["band"] == "g")

        # Don't care if mounted, swap it
        detailer = detailers.BandPickToODetailer(
            band_start="g", band_end="r", distance_limit=10.0, check_mounted=False
        )

        out_obs = detailer(obs, conditions, target_o_o=too_event)
        assert np.all(out_obs["band"] == "r")

    def test_trackinginfo(self):
        obs = ObservationArray(2)
        conditions = Conditions()

        # Check basic function
        det = detailers.TrackingInfoDetailer(
            target_name="", observation_reason="pairs_12", science_program="BLOCK-T"
        )

        out_obs = det(obs, conditions)
        for o in out_obs:
            self.assertEqual(o["observation_reason"], "pairs_12")
            self.assertEqual(o["science_program"], "BLOCK-T")
            self.assertEqual(o["target_name"], "")

        # Check that it replaces spaces with underscores
        obs = ObservationArray(2)
        det = detailers.TrackingInfoDetailer(
            target_name="", observation_reason="pairs space", science_program="BLOCK-T"
        )

        out_obs = det(obs, conditions)
        for o in out_obs:
            self.assertEqual(o["observation_reason"], "pairs_space")
            self.assertEqual(o["science_program"], "BLOCK-T")
            self.assertEqual(o["target_name"], "")

        # Check that it doesn't overwrite existing data
        det = detailers.TrackingInfoDetailer(
            target_name="", observation_reason="pairs_12", science_program="BLOCK-T"
        )
        out_obs = det(out_obs, conditions)
        for o in out_obs:
            self.assertEqual(o["observation_reason"], "pairs_space")

    def test_band_sort(self):
        """Test BandSortDetailer"""

        detailer = detailers.BandSortDetailer()
        conditions = Conditions()

        conditions.current_band = "i"

        obs = ObservationArray(2)
        obs["band"][0] = "r"
        obs["band"][1] = "z"

        # Default settings should not change order
        obs_out = detailer(obs, conditions)
        assert (obs_out["band"][0] == "r") & (obs_out["band"][1] == "z")

        # Swap, make sure still doesn't change order
        obs = ObservationArray(2)
        obs["band"][0] = "z"
        obs["band"][1] = "r"

        # Default settings should not change order
        obs_out = detailer(obs, conditions)
        assert (obs_out["band"][1] == "r") & (obs_out["band"][0] == "z")

        # If r is loaded, should always swap to having r first
        conditions.current_band = "r"

        obs = ObservationArray(2)
        obs["band"][0] = "r"
        obs["band"][1] = "z"

        # Default settings should not change order
        obs_out = detailer(obs, conditions)
        assert (obs_out["band"][0] == "r") & (obs_out["band"][1] == "z")

        # Swap, make sure still doesn't change order
        obs = ObservationArray(2)
        obs["band"][0] = "z"
        obs["band"][1] = "r"

        # Default settings should not change order
        obs_out = detailer(obs, conditions)
        assert (obs_out["band"][0] == "r") & (obs_out["band"][1] == "z")

        # If we want a specified order, ignore what's loaded
        detailer = detailers.BandSortDetailer(desired_band_order="yzi", loaded_first=False)
        conditions = Conditions()
        conditions.current_band = "i"
        obs = ObservationArray(3)
        obs["band"][0] = "i"
        obs["band"][1] = "y"
        obs["band"][2] = "z"

        obs_out = detailer(obs, conditions)

        assert obs_out["band"][0] == "y"
        assert obs_out["band"][1] == "z"
        assert obs_out["band"][2] == "i"

        # Same, but now do bump up band if loaded
        detailer = detailers.BandSortDetailer(desired_band_order="yzi", loaded_first=True)
        conditions = Conditions()
        conditions.current_band = "i"
        obs = ObservationArray(3)
        obs["band"][0] = "i"
        obs["band"][1] = "y"
        obs["band"][2] = "z"

        obs_out = detailer(obs, conditions)

        assert obs_out["band"][0] == "i"
        assert obs_out["band"][1] == "y"
        assert obs_out["band"][2] == "z"

    def test_splitdither(self):
        obs = ObservationArray(5)
        input_notes = ["survey_a", "survey_b", "survey_a", "survey_b", "survey_a"]
        for i in range(len(obs)):
            obs[i]["scheduler_note"] = input_notes[i]

        conditions = Conditions()

        # This is more likely used to split between dither patterns,
        # But for easy test let's just assign different target_names.
        det_a = detailers.TrackingInfoDetailer(target_name="a")
        det_b = detailers.TrackingInfoDetailer(target_name="b")

        det = detailers.SplitDetailer(det_a, det_b, split_str="_b")
        out_obs = det(obs, conditions)

        for o in out_obs:
            assert o["scheduler_note"].split("_")[-1] == o["target_name"]

    def test_region_label(self):

        for detailer_start in [detailers.LabelRegionDetailer, detailers.LabelRegionsAndDDFs]:
            nside = 32
            sky = CurrentAreaMap(nside=nside)
            footprints, labels = sky.return_maps()

            obs = ObservationArray(3)
            obs["RA"] = np.radians(0.0)
            obs["dec"] = np.radians(-20.0)

            obs["target_name"] = "dummy_start"

            detailer = detailer_start(label_array=labels)

            output = detailer(obs, None)
            for res in output:
                assert "dummy" in res["target_name"]
                assert "lowdust" in res["target_name"]

        # Test that we clobber
        detailer = detailers.LabelRegionDetailer(labels, append=False)
        output = detailer(obs, None)
        for res in output:
            assert "dummy" not in res["target_name"]
            assert "lowdust" in res["target_name"]

        # Test that we can have multiple labels
        obs = ObservationArray(3)
        obs["RA"] = np.radians(0.0)
        obs["dec"] = np.radians(3.0)
        detailer = detailers.LabelRegionDetailer(labels)
        output = detailer(obs, None)
        for res in output:
            assert "dummy" not in res["target_name"]
            assert "lowdust" in res["target_name"]
            assert "nes" in res["target_name"]

    def test_euclid_dither(self):
        detailer = detailers.EuclidDitherDetailer(per_night=True)

        ra_a, dec_a, ra_b, dec_b = detailer._generate_offsets(100, 2)

        # Going per night, should only be one offset
        assert np.size(ra_a) == 1

        # if we call again, should be the same
        ra_ap, dec_ap, ra_bp, dec_bp = detailer._generate_offsets(100, 2)

        assert ra_a == ra_ap
        assert dec_a == dec_ap
        assert ra_b == ra_bp
        assert dec_b == dec_bp

        # Now want a unique position
        detailer = detailers.EuclidDitherDetailer(per_night=False)

        ra_a, dec_a, ra_b, dec_b = detailer._generate_offsets(100, 2)

        assert np.size(ra_a) == 100

        # If those get executed, they get added to the detailer
        obs = ObservationArray(n=100)
        obs["night"] = 2
        detailer.add_observations_array(obs, None)

        # call again should result in different values
        ra_ap, dec_ap, ra_bp, dec_bp = detailer._generate_offsets(100, 2)
        assert np.all(np.not_equal(ra_a, ra_ap))

    def test_dither(self):
        detailer = detailers.DitherDetailer(per_night=True)

        ra_a, dec_a = detailer._generate_offsets(100, 2)
        # Going per night, should only be one offset
        assert np.size(np.unique(ra_a)) == 1

        # if we call again, should be the same
        ra_ap, dec_ap = detailer._generate_offsets(100, 2)

        assert np.all(ra_a == ra_ap)
        assert np.all(dec_a == dec_ap)

        # Now want a unique position
        detailer = detailers.DitherDetailer(per_night=False)

        ra_a, dec_a = detailer._generate_offsets(100, 2)

        assert np.size(np.unique(ra_a)) == 100

        # If those get executed, they get added to the detailer
        obs = ObservationArray(n=100)
        obs["night"] = 2
        detailer.add_observations_array(obs, None)

        # call again should result in different values
        ra_ap, dec_ap = detailer._generate_offsets(100, 2)

        assert np.all(np.not_equal(ra_a, ra_ap))

        # Check that negative nights are ok
        detailer = detailers.DitherDetailer(per_night=True)
        detailer._new_ang_rad(-10)
        ang1 = detailer.angles.copy()
        detailer._new_ang_rad(10)
        ang2 = detailer.angles.copy()

        assert np.all(ang1 != ang2)

    def test_truncate(self):
        """Test we truncate observations before twilight"""
        detailer = detailers.TruncatePreTwiDetailer()
        observation_array = ObservationArray(n=100)
        n1 = np.size(observation_array)

        observation_array["band"] = "r"
        observation_array["exptime"] = 30.0

        conditions = Conditions(mjd=3, survey_start_mjd=0)
        conditions.current_band = "r"
        # Have a ton of time, no obs should be cut
        conditions.sun_n18_rising = 3.5
        obs_out = detailer(observation_array, conditions)

        assert np.size(obs_out) == n1

        # cut down to not enough time
        conditions.sun_n18_rising = 3.016

        obs_out = detailer(observation_array, conditions)
        n2 = np.size(obs_out)

        # Now if we are in a different filter, should
        # Have to lose a few more, but not all
        conditions.current_band = "b"
        obs_out = detailer(observation_array, conditions)

        n3 = np.size(obs_out)

        assert n1 > n2

        assert n2 > n3

        assert n3 > 0

    def test_rollband(self):

        orig_order = np.array(["u", "g", "r"])

        observation_array = ObservationArray(n=3)
        observation_array["band"] = ["u", "g", "r"]

        # Test where no change should happen
        detailer = detailers.RollBandMatchDetailer()
        conditions = Conditions()
        conditions.current_band = "u"
        o1 = detailer(observation_array, conditions)
        for bn, ob in zip(orig_order, o1["band"]):
            assert bn == ob

        # Test where things should roll
        conditions.current_band = "r"
        o2 = detailer(observation_array, conditions)

        assert o2["band"][0] == "r"

        for bn in orig_order:
            assert bn in o2["band"]


if __name__ == "__main__":
    unittest.main()



__all__ = ("Site",)

import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation


class LsstSiteParameters:
    """
    This is a struct containing the LSST site parameters as defined in

    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-30

    (accessed on 4 January 2016)

    Updated from:
    https://rubinobs.atlassian.net/wiki/spaces/LTS/
    pages/50084908/Rubin+Telescopes+
    Latitude+and+Longitude+Coordinates?
    force_transition=c54e56dd-7d7b-42c2-b5e5-79aa1c1bc777

    This class only exists for initializing Site with LSST parametervalues.
    Users should not be accessing this class directly.
    """

    def __init__(self):
        self.longitude = -70.74941  # in degrees
        self.latitude = -30.244628  # in degrees
        self.height = 2650.0  # in meters
        self.temperature = 11.5  # in centigrade
        self.pressure = 750.0  # in millibars
        self.humidity = 0.4  # scale 0-1
        self.lapse_rate = 0.0065  # in Kelvin per meter
        # the lapse rate was not specified by LSE-30;
        # 0.0065 K/m appears to be the "standard" value
        # see, for example
        # http://mnras.oxfordjournals.org/content/365/4/1235.full


class AuxTelParameters:
    """
    Struct with AuxTel default parameters

    Values from:
    https://rubinobs.atlassian.net/wiki/spaces/LTS/
    pages/50084908/Rubin+Telescopes+
    Latitude+and+Longitude+Coordinates?
    force_transition=c54e56dd-7d7b-42c2-b5e5-79aa1c1bc777

    """

    def __init__(self):
        self.longitude = -70.747686  # in degrees
        self.latitude = -30.244789  # in degrees
        # Setting these the same as LSSTSiteParameters
        self.height = 2650.0  # in meters
        self.temperature = 11.5  # in centigrade
        self.pressure = 750.0  # in millibars
        self.humidity = 0.4  # scale 0-1
        self.lapse_rate = 0.0065  # in Kelvin per meter


class Site:
    """
    This class will store site information for use in Catalog objects.

    Defaults values are LSST site values taken from the Observatory
    System Specification document
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-30
    on 4 January 2016

    Parameters
    ----------
    name : `str`, opt
        The name of the observatory. Set to 'LSST' for other parameters
        to default to LSST values. Set to 'AuxTel' for
        auxtel values.
    longitude : `float`, opt
        Longitude of the site in degrees.
    latitude : `float`, opt
        Latitude of the site in degrees.
    height : `float`, opt
        Height of the site in meters.
    temperature : `float`, opt
        Mean temperature in Centigrade
    pressure : `float`, opt
        Pressure for the site in millibars.
    humidity : `float`, opt
        Relative humidity (range 0-1).
    lapse_rate : `float`, opt
        Change in temperature in Kelvins per meter
    """

    def __init__(
        self,
        name=None,
        longitude=None,
        latitude=None,
        height=None,
        temperature=None,
        pressure=None,
        humidity=None,
        lapse_rate=None,
    ):
        default_params = None
        self._name = name
        if self._name == "LSST":
            default_params = LsstSiteParameters()

        if self.name == "AuxTel":
            default_params = AuxTelParameters()

        if default_params is not None:
            if longitude is None:
                longitude = default_params.longitude

            if latitude is None:
                latitude = default_params.latitude

            if height is None:
                height = default_params.height

            if temperature is None:
                temperature = default_params.temperature

            if pressure is None:
                pressure = default_params.pressure

            if humidity is None:
                humidity = default_params.humidity

            if lapse_rate is None:
                lapse_rate = default_params.lapse_rate

        if longitude is not None:
            self._longitude_rad = np.radians(longitude)
        else:
            self._longitude_rad = None

        if latitude is not None:
            self._latitude_rad = np.radians(latitude)
        else:
            self._latitude_rad = None

        self._longitude_deg = longitude
        self._latitude_deg = latitude
        self._height = height
        self._pressure = pressure

        if temperature is not None:
            self._temperature_kelvin = temperature + 273.15  # in Kelvin
        else:
            self._temperature_kelvin = None

        self._temperature_centigrade = temperature
        self._humidity = humidity
        self._lapse_rate = lapse_rate

        # Go through all the attributes of this Site.
        # Raise a warning if any are None so that the user
        # is not surprised when some use of this Site fails
        # because something that should have beena a float
        # is NoneType
        list_of_nones = []
        if self.longitude is None or self.longitude_rad is None:
            if self.longitude_rad is not None:
                raise RuntimeError("in Site: longitude is None but longitude_rad is not")
            if self.longitude is not None:
                raise RuntimeError("in Site: longitude_rad is None but longitude is not")
            list_of_nones.append("longitude")

        if self.latitude is None or self.latitude_rad is None:
            if self.latitude_rad is not None:
                raise RuntimeError("in Site: latitude is None but latitude_rad is not")
            if self.latitude is not None:
                raise RuntimeError("in Site: latitude_rad is None but latitude is not")
            list_of_nones.append("latitude")

        if self.temperature is None or self.temperature_kelvin is None:
            if self.temperature is not None:
                raise RuntimeError("in Site: temperature_kelvin is None but temperature is not")
            if self.temperature_kelvin is not None:
                raise RuntimeError("in Site: temperature is None but temperature_kelvin is not")
            list_of_nones.append("temperature")

        if self.height is None:
            list_of_nones.append("height")

        if self.pressure is None:
            list_of_nones.append("pressure")

        if self.humidity is None:
            list_of_nones.append("humidity")

        if self.lapse_rate is None:
            list_of_nones.append("lapse_rate")

        if len(list_of_nones) != 0:
            msg = "The following attributes of your Site were None:\n"
            for name in list_of_nones:
                msg += "%s\n" % name
            msg += "If you want these to just default to LSST values,\n"
            msg += "instantiate your Site with name='LSST'"
            warnings.warn(msg)

    def to_earth_location(self):
        return EarthLocation.from_geodetic(
            lon=self.longitude * u.deg, lat=self.latitude * u.deg, height=self.height * u.meter
        )

    def __eq__(self, other):
        for param in self.__dict__:
            if param not in other.__dict__:
                return False
            if self.__dict__[param] != other.__dict__[param]:
                return False

        for param in other.__dict__:
            if param not in self.__dict__:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def name(self):
        """
        observatory name
        """
        return self._name

    @property
    def longitude_rad(self):
        """
        observatory longitude in radians
        """
        return self._longitude_rad

    @property
    def longitude(self):
        """
        observatory longitude in degrees
        """
        return self._longitude_deg

    @property
    def latitude_rad(self):
        """
        observatory latitude in radians
        """
        return self._latitude_rad

    @property
    def latitude(self):
        """
        observatory latitude in degrees
        """
        return self._latitude_deg

    @property
    def temperature(self):
        """
        mean temperature in centigrade
        """
        return self._temperature_centigrade

    @property
    def temperature_kelvin(self):
        """
        mean temperature in Kelvin
        """
        return self._temperature_kelvin

    @property
    def height(self):
        """
        height in meters
        """
        return self._height

    @property
    def pressure(self):
        """
        mean pressure in millibars
        """
        return self._pressure

    @property
    def humidity(self):
        """
        mean humidity in the range 0-1
        """
        return self._humidity

    @property
    def lapse_rate(self):
        """
        temperature lapse rate (in Kelvin per meter)
        """
        return self._lapse_rate



import unittest

from astropy.time import Time, TimeDelta

from rubin_scheduler.site_models import DowntimeModel, ScheduledDowntimeData, UnscheduledDowntimeData


class TestDowntimeModel(unittest.TestCase):
    def test_call(self):
        # Check the calculation from fwhm_500 to fwhm_eff/fwhm_geom.
        # Use simple effective wavelengths and airmass values.
        downtime_model = DowntimeModel()
        t = Time("2022-10-01")
        sched = ScheduledDowntimeData(t)
        sched.read_data()
        unsched = UnscheduledDowntimeData(t)
        unsched.make_data()
        efd_data = {"unscheduled_downtimes": unsched(), "scheduled_downtimes": sched()}
        # Set time to within first scheduled downtime.
        t_now = sched.downtime[0]["start"] + TimeDelta(0.5, format="jd")
        target_dict = {"time": t_now}
        dt_status = downtime_model(efd_data, target_dict)
        # Expect return dict of : {'status': status, 'end': end_down,
        # 'next': next_sched['start']}
        # Check keys
        for k in ("status", "end", "next"):
            self.assertTrue(k in dt_status)
        # downtime status is "True" if system is down.
        self.assertEqual(True, dt_status["status"])
        self.assertEqual(dt_status["end"], sched.downtime[0]["end"])
        self.assertEqual(dt_status["next"], sched.downtime[1]["start"])


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "ObservationArray",
    "ScheduledObservationArray",
    "backfill_filter_from_band",
)

import numpy as np

HANDLED_FUNCTIONS = {}


def backfill_filter_from_band(observation_array):
    """Simple utility to backfill the `band` values in an ObservationArray
    if not otherwise present.

    Parameters
    ----------
    observation_array : ObservationArray
        The observation array to ensure that 'band' is present in, if
        only physical filtername was present.

    Returns
    -------
    observation_array : ObservationArray
        The same observation_array, but with `band` added where it was missing.

    Notes
    -----
    This is primarily a utility to provide backwards compatibility for
    scheduler configurations where ObservationArrays are input (such as
    for FieldSurvey), but only have `filter` present, instead of also `band`.
    For the current setup for comcam and lsstcam, the stripping of
    "_0123456789" is sufficient, in case the configuration used physical
    filtername --
    physical filtername <> band includes "r_03" <> "r", for example.
    """
    missing_band = np.where(observation_array["band"] == "")
    band = np.char.rstrip(observation_array["filter"][missing_band], chars="_0123456789")
    observation_array["band"][missing_band] = band
    return observation_array


class ObservationArray(np.ndarray):
    """Class to work as an array of observations

    Parameters
    ----------
    n : `int`
        Size of array to return. Default 1.

    The numpy fields have the following labels.

    RA : `float`
       The Right Acension of the observation (center of the field)
       (Radians)
    dec : `float`
       Declination of the observation (Radians)
    mjd : `float`
       Modified Julian Date at the start of the observation
       (time shutter opens)
    exptime : `float`
       Total exposure time of the visit (seconds)
    band : `str`
        The band used. Should be one of u, g, r, i, z, y.
    filter : `str`
        The physical filter name.
    rotSkyPos : `float`
        The rotation angle of the camera relative to the sky E of N
        (Radians). Will be ignored if rotTelPos is finite.
        If rotSkyPos is set to NaN, rotSkyPos_desired is used.
    rotTelPos : `float`
        The rotation angle of the camera relative to the telescope
        (radians). Set to np.nan to force rotSkyPos to be used.
    rotSkyPos_desired : `float`
        If both rotSkyPos and rotTelPos are None/NaN, then
        rotSkyPos_desired (radians) is used. If rotSkyPos_desired
        results in a valid rotTelPos, rotSkyPos is set to
        rotSkyPos_desired. If rotSkyPos and rotTelPos are both NaN,
        and rotSkyPos_desired results in an out of range value for the
        camera rotator, then rotTelPos_backup is used.
    rotTelPos_backup : `float`
        Rotation angle of the camera relative to the telescope (radians).
        Only used as a last resort if rotSkyPos and rotTelPos are set
        to NaN and rotSkyPos_desired results in an out of range rotator
        value.
    nexp : `int`
        Number of exposures in the visit.
    flush_by_mjd : `float`
        If we hit this MJD, we should flush the queue and refill it.
    scheduler_note : `str` (optional)
        Usually good to set the note field so one knows which survey
        object generated the observation.
    target_name : `str` (optional)
        A note about what target is being observed.
        This maps to target_name in the ConsDB.
        Generally would be used to identify DD, ToO or special targets.
    science_program : `str` (optional)
        Science program being executed.
        This maps to science_program in the ConsDB, although can
        be overwritten in JSON BLOCK.
        Generally would be used to identify a particular program for DM.
    observation_reason : `str` (optional)
        General 'reason' for observation, for DM purposes.
        (for scheduler purposes, use `scheduler_note`).
        This maps to observation_reason in the ConsDB, although could
        be overwritten in JSON BLOCK.
        Most likely this is just "science" or "FBS" when using the FBS.
    cloud_extinction : float
        Cloud extinction that was applied by the model observatory (mags).

    Notes
    -----

    On the camera rotator angle. Order of priority goes:
    rotTelPos > rotSkyPos > rotSkyPos_desired > rotTelPos_backup
    where if rotTelPos is NaN, it checks rotSkyPos. If rotSkyPos is set,
    but not at an accessible rotTelPos, the observation will fail.
    If rotSkyPos is NaN, then rotSkyPos_desired is used. If
    rotSkyPos_desired is at an inaccessbile rotTelPos, the observation
    does not fail, but falls back to the value in rotTelPos_backup.

    Lots of additional fields that get filled in by the model observatory
    when the observation is completed.
    See documentation at:
    https://rubin-scheduler.lsst.io/output_schema.html

    """

    def __new__(cls, n=1):
        dtypes = [
            ("ID", int),
            ("RA", float),
            ("dec", float),
            ("mjd", float),
            ("flush_by_mjd", float),
            ("exptime", float),
            ("band", "U40"),
            ("filter", "U40"),
            ("rotSkyPos", float),
            ("rotSkyPos_desired", float),
            ("nexp", int),
            ("airmass", float),
            ("FWHM_500", float),
            ("FWHMeff", float),
            ("FWHM_geometric", float),
            ("skybrightness", float),
            ("night", int),
            ("slewtime", float),
            ("visittime", float),
            ("slewdist", float),
            ("fivesigmadepth", float),
            ("alt", float),
            ("az", float),
            ("pa", float),
            ("pseudo_pa", float),
            ("clouds", float),
            ("moonAlt", float),
            ("sunAlt", float),
            ("scheduler_note", "U40"),
            ("target_name", "U40"),
            ("target_id", int),
            ("lmst", float),
            ("rotTelPos", float),
            ("rotTelPos_backup", float),
            ("moonAz", float),
            ("sunAz", float),
            ("sunRA", float),
            ("sunDec", float),
            ("moonRA", float),
            ("moonDec", float),
            ("moonDist", float),
            ("solarElong", float),
            ("moonPhase", float),
            ("cummTelAz", float),
            ("observation_reason", "U40"),
            ("science_program", "U40"),
            ("cloud_extinction", float),
            # TODO : Remove this hack which is for use with ts_scheduler
            # version <=v2.3 .. remove ts_scheduler actually drops "note".
            ("note", "U40"),
        ]
        obj = np.zeros(n, dtype=dtypes).view(cls)
        return obj

    def tolist(self):
        """Convert to a list of 1-element arrays"""
        obs_list = []
        for obs in self:
            new_obs = self.__class__(n=1)
            new_obs[0] = obs
            obs_list.append(new_obs)

        return obs_list

    def __array_function__(self, func, types, args, kwargs):
        # If we want "standard numpy behavior",
        # convert any ObservationArray to ndarray views
        if func not in HANDLED_FUNCTIONS:
            new_args = []
            for arg in args:
                if issubclass(arg.__class__, ObservationArray):
                    new_args.append(arg.view(np.ndarray))
                else:
                    new_args.append(arg)
            return func(*new_args, **kwargs)
        if not all(issubclass(t, ObservationArray) for t in types):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)


def implements(numpy_function):
    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@implements(np.concatenate)
def concatenate(arrays):
    result = arrays[0].__class__(n=sum(len(a) for a in arrays))
    return np.concatenate([np.asarray(a) for a in arrays], out=result)


@implements(np.in1d)
def in1d(ar1, ar2, **kwargs):
    # Just here to get rid of DeprecationWarning. Can probably
    # be removed after a future numpy update.
    return np.isin(ar1, ar2, **kwargs)


class ScheduledObservationArray(ObservationArray):
    """Make an array to hold pre-scheduling observations

    Note
    ----
    mjd_tol : `float`
        The tolerance on how early an observation can execute (days).
        Observation will be considered valid to attempt
        when mjd-mjd_tol < current MJD < flush_by_mjd (and other
        conditions below pass)
    dist_tol : `float`
        The angular distance an observation can be away from the
        specified RA,Dec and still count as completing the observation
        (radians).
    alt_min : `float`
        The minimum altitude to consider executing the observation
        (radians).
    alt_max : `float`
        The maximuim altitude to try observing (radians).
    HA_max : `float`
        Hour angle limit. Constraint is such that for hour angle
        running from 0 to 24 hours, the target RA,Dec must be greather
        than HA_max and less than HA_min. Set HA_max to 0 for no
        limit. (hours)
    HA_min : `float`
        Hour angle limit. Constraint is such that for hour angle
        running from 0 to 24 hours, the target RA,Dec must be greather
        than HA_max and less than HA_min. Set HA_min to 24 for
        no limit. (hours)
    sun_alt_max : `float`
        The sun must be below sun_alt_max to execute. (radians)
    moon_min_distance : `float`
        The minimum distance to demand the moon should be away (radians)
    observed : `bool`
        If set to True, scheduler will probably consider this a
        completed observation and never attempt it.

    """

    def __new__(cls, n=1):
        # Standard things from the usual observations
        dtypes1 = [
            ("ID", int),
            ("RA", float),
            ("dec", float),
            ("mjd", float),
            ("flush_by_mjd", float),
            ("exptime", float),
            ("band", "U1"),
            ("filter", "U40"),
            ("rotSkyPos", float),
            ("rotTelPos", float),
            ("rotTelPos_backup", float),
            ("rotSkyPos_desired", float),
            ("nexp", int),
            ("scheduler_note", "U40"),
            ("target_name", "U40"),
            ("science_program", "U40"),
            ("observation_reason", "U40"),
        ]

        # New things not in standard ObservationArray
        dtype2 = [
            ("mjd_tol", float),
            ("dist_tol", float),
            ("alt_min", float),
            ("alt_max", float),
            ("HA_max", float),
            ("HA_min", float),
            ("sun_alt_max", float),
            ("moon_min_distance", float),
            ("observed", bool),
        ]

        obj = np.zeros(n, dtype=dtypes1 + dtype2).view(cls)
        return obj

    def to_observation_array(self):
        """Convert the scheduled observation to a
        Regular ObservationArray
        """
        result = ObservationArray(n=self.size)
        in_common = np.intersect1d(self.dtype.names, result.dtype.names)
        for key in in_common:
            result[key] = self[key]
        return result



import os
import pathlib
import unittest

import healpy as hp
import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.example import example_scheduler, run_sched
from rubin_scheduler.scheduler.model_observatory import KinemModel, ModelObservatory
from rubin_scheduler.scheduler.utils import (
    ObservationArray,
    ScheduledObservationArray,
    SchemaConverter,
    ecliptic_area,
    make_rolling_footprints,
    restore_scheduler,
    run_info_table,
    season_calc,
)
from rubin_scheduler.utils import SURVEY_START_MJD


class TestUtils(unittest.TestCase):
    def tearDownClass():
        pathlib.Path("temp.sqlite").unlink(missing_ok=True)

    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_nside(self):
        """Test the example scheduler can be set to different nsides."""
        mjd_start = SURVEY_START_MJD
        _ = example_scheduler(mjd_start=mjd_start, nside=64)
        _ = example_scheduler(mjd_start=mjd_start, nside=8)

    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_start_of_night_example(self):
        """Test example scheduler and sim_runner having mis-matched
        start dates."""
        mjd_start = SURVEY_START_MJD
        scheduler = example_scheduler(mjd_start=mjd_start)
        observatory, scheduler, observations = run_sched(
            scheduler, mjd_start=mjd_start - 0.5, survey_length=5
        )

    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_altaz_limit(self):
        """Test that setting some azimuth limits via different approaches
        for the AltAzShadowMaskBasisFunction works"""
        mjd_start = SURVEY_START_MJD
        scheduler = example_scheduler(mjd_start=mjd_start)

        # Constrain the pointings available with the telescope mount
        # This test needs to run in the north, not the south
        km = KinemModel(mjd0=mjd_start)
        az_min = 260
        az_max = 110
        km.setup_telescope(azimuth_minpos=az_min, azimuth_maxpos=az_max)
        mo = ModelObservatory(mjd=mjd_start, mjd_start=mjd_start, kinem_model=km)
        try:
            mo, scheduler, observations = sim_runner(
                mo,
                scheduler,
                sim_duration=3.0,
                verbose=False,
                filename=None,
            )
        except RuntimeError as e:
            print("Failed in az limits test")
            raise e

        az = np.degrees(observations["az"])
        forbidden = np.where((az > az_max) & (az < az_min))[0]
        assert forbidden.size == 0

        km = KinemModel(mjd0=mjd_start)
        alt_min = 40
        alt_max = 70
        km.setup_telescope(altitude_minpos=alt_min, altitude_maxpos=alt_max)
        mo = ModelObservatory(
            mjd=observations[-1]["mjd"],
            mjd_start=mjd_start,
            kinem_model=km,
            downtimes="ideal",
            cloud_data="ideal",
        )
        scheduler.flush_queue()
        try:
            mo, scheduler, observations = sim_runner(
                mo,
                scheduler,
                sim_duration=3.0,
                verbose=False,
                filename=None,
            )
        except RuntimeError as e:
            print("Failed in altitude test")
            raise e
        alt = np.degrees(observations["alt"])
        n_forbidden = np.size(np.where((alt > alt_max) & (alt < alt_min))[0])
        assert n_forbidden == 0

        # Constrain the pointings available with
        # limits for the model observatory
        km = KinemModel(mjd0=mjd_start)
        az_min = 260
        az_max = 110
        mo = ModelObservatory(
            mjd=observations[-1]["mjd"],
            mjd_start=mjd_start,
            kinem_model=km,
            sky_az_limits=[
                [az_min, az_max],
            ],
        )
        try:
            mo, scheduler, observations = sim_runner(
                mo,
                scheduler,
                sim_duration=3.0,
                verbose=False,
                filename=None,
            )
        except RuntimeError as e:
            print("Failed in az limits test")
            raise e

        az = np.degrees(observations["az"])
        forbidden = np.where((az > az_max) & (az < az_min))[0]
        # Some visits will occur in this forbidden range, due to
        # some complications about the shadow mask and pair completion
        # Since this is like a "preference" when set with sky_az_limits,
        # should allow some to occur in this set of forbidden azimuths
        second_pairs = [
            scheduler_note
            for scheduler_note in observations[forbidden]["scheduler_note"]
            if "pair" in scheduler_note and ", b" in scheduler_note
        ]
        assert forbidden.size == len(second_pairs)
        assert np.all(((az[forbidden] - az_min) % (az_max - az_min)) < 3)

        km = KinemModel(mjd0=mjd_start)
        mo = ModelObservatory(
            mjd=observations[-1]["mjd"],
            mjd_start=mjd_start,
            kinem_model=km,
            sky_alt_limits=[
                [alt_min, alt_max],
            ],
        )
        scheduler.flush_queue()
        try:
            mo, scheduler, observations = sim_runner(
                mo,
                scheduler,
                sim_duration=3.0,
                verbose=False,
                filename=None,
            )
        except RuntimeError as e:
            print("Failed in altitude test")
            raise e
        alt = np.degrees(observations["alt"])
        forbidden = np.where((alt > alt_max) & (alt < alt_min))[0]
        assert forbidden.size == 0

    def test_restore(self):
        """Test we can restore a scheduler properly"""
        # MJD set so it's in test data range
        mjd_start = SURVEY_START_MJD
        n_visit_limit = 3000

        scheduler = example_scheduler(mjd_start=mjd_start)

        mo = ModelObservatory(mjd_start=mjd_start, downtimes="ideal", cloud_data="ideal")
        # Never load too many nights of sky
        mo.sky_model.load_length = 10.0
        mo, scheduler, observations = sim_runner(
            mo,
            scheduler,
            sim_duration=30.0,
            verbose=False,
            filename=None,
            n_visit_limit=n_visit_limit,
        )

        # Won't be exact if we restart in the middle of a blob sequence
        # since the queue isn't reconstructed.
        # Also, any scripted observations that get generated
        # during the night (e.g., long gaps observations) will get lost,
        # so need to restart on a new night to ensure identical results.

        nd = np.zeros(observations.size)
        nd[1:] = np.diff(observations["night"])

        break_indx = np.min(np.where((observations["ID"] >= n_visit_limit / 2.0) & (nd != 0))[0])
        new_n_limit = n_visit_limit - break_indx

        new_mo = ModelObservatory(mjd_start=mjd_start, downtimes="ideal", cloud_data="ideal")
        # Never load too much sky
        new_mo.sky_model.load_length = 10.0
        new_sched = example_scheduler(mjd_start=mjd_start)

        # Restore some of the observations
        new_sched, new_mo = restore_scheduler(break_indx - 1, new_sched, new_mo, observations, fast=False)

        # Simulate ahead and confirm that it behaves the same
        # as running straight through
        new_mo, new_sched, new_obs = sim_runner(
            new_mo,
            new_sched,
            sim_duration=20.0,
            verbose=False,
            filename=None,
            n_visit_limit=new_n_limit,
        )

        # Check that observations taken after restart match those from
        # before Jenkins can be bad at comparing things, so if it thinks
        # they aren't the same, check column-by-column to double check
        if not np.all(new_obs == observations[break_indx:]):
            names = new_obs.dtype.names
            for name in names:
                # If it's a string
                if new_obs[name].dtype == "<U40":
                    assert np.all(new_obs[name] == observations[break_indx:][name])
                # Otherwise should be number-like
                else:
                    assert np.allclose(new_obs[name], observations[break_indx:][name])
        # Didn't need to go by column, the observations after restart
        # match the ones that were taken all at once.
        else:
            assert np.all(new_obs == observations[break_indx:])

        # And again, but this time using the fast array restore
        new_mo = ModelObservatory(mjd_start=mjd_start, downtimes="ideal", cloud_data="ideal")
        new_mo.sky_model.load_length = 10.0
        new_sched = example_scheduler(mjd_start=mjd_start)
        new_sched, new_mo = restore_scheduler(break_indx - 1, new_sched, new_mo, observations, fast=True)
        # Simulate ahead and confirm that it behaves the same as
        # running straight through
        new_mo, new_sched, new_obs_fast = sim_runner(
            new_mo,
            new_sched,
            sim_duration=20.0,
            verbose=False,
            filename=None,
            n_visit_limit=new_n_limit,
        )

        # Check that observations taken after restart match those
        # from before Jenkins can be bad at comparing things, so if
        # it thinks they aren't the same, check column-by-column to
        # double check
        if not np.all(new_obs_fast == observations[break_indx:]):
            names = new_obs_fast.dtype.names
            for name in names:
                # If it's a string
                if new_obs_fast[name].dtype == "<U40":
                    assert np.all(new_obs_fast[name] == observations[break_indx:][name])
                # Otherwise should be number-like
                else:
                    assert np.allclose(new_obs_fast[name], observations[break_indx:][name])
        # Didn't need to go by column, the observations after restart
        # match the ones that were taken all at once.
        else:
            assert np.all(new_obs_fast == observations[break_indx:])

    def test_season(self):
        """
        Test that the season utils work as intended
        """
        night = 365.25 * 3.5
        plain = season_calc(night)
        assert plain == 3

        mod2 = season_calc(night, modulo=2)
        assert mod2 == 1

        mod3 = season_calc(night, modulo=3)
        assert mod3 == 0

        mod3 = season_calc(night, modulo=3, max_season=2)
        assert mod3 == -1

        mod3 = season_calc(night, modulo=3, max_season=2, offset=-365.25 * 2)
        assert mod3 == 1

        mod3 = season_calc(night, modulo=3, max_season=2, offset=-365.25 * 10)
        assert mod3 == -1

        mod3 = season_calc(night, modulo=3, offset=-365.25 * 10)
        assert mod3 == -1

    def test_observation_array(self):
        # Check we can convert ObservationArrays to lists and back
        n = 20
        obs = ObservationArray(n=n)
        obs["ID"] = np.arange(n)

        obs_list = obs.tolist()
        assert len(obs_list[0]) == 1
        assert len(obs_list) == n

        back = np.concatenate(obs_list)

        assert np.array_equal(back, obs)

        # and scheduled arrays
        obs = ScheduledObservationArray(n=n)
        obs["ID"] = np.arange(n)

        obs_list = obs.tolist()
        assert len(obs_list[0]) == 1
        assert len(obs_list) == n

        back = np.concatenate(obs_list)

        assert np.array_equal(back, obs)

    def test_schema_convert(self):
        sc = SchemaConverter()

        n = 20
        obs = ObservationArray(n=n)
        obs["ID"] = np.arange(n)

        # check that we can write observations to a database
        sc.obs2opsim(obs, filename="temp.sqlite")

        # change the ID and write it again, should append
        obs["ID"] = np.arange(n) + n + 1
        sc.obs2opsim(obs, filename="temp.sqlite")

        read_in = sc.opsim2obs("temp.sqlite")

        assert len(read_in) == 2 * n

    def test_run_info_table(self):
        """Test run_info_table gets information"""
        observatory = ModelObservatory(nside=8, mjd_start=SURVEY_START_MJD)
        version_info = run_info_table(observatory)
        # Make a minimal set of keys that ought to be in the info table
        # Update these if the value they're stored as changes
        # (either from run_info_table or observatory.info)
        need_keys = [
            "rubin_scheduler.__version__",
            "hostname",
            "Date, ymd",
            "site_models",
            "skybrightness_pre",
        ]
        have_keys = list(version_info["Parameter"])
        for k in need_keys:
            self.assertTrue(k in have_keys)

    def test_ecliptic_area(self):

        result = ecliptic_area()
        assert np.sum(result) > 0

        result = ecliptic_area(nside=128, dist_to_eclip=50.0, dec_max=-10.0)
        assert np.sum(result) > 0

    def test_make_rolling_footprints_uniform(self):
        """Test that we can make a uniform folling footprint"""

        # utility function to get sun ra at a given mjd
        from astropy.coordinates import EarthLocation, get_sun
        from astropy.time import Time

        def _get_sun_ra_at_mjd(mjd):
            t = Time(
                mjd,
                format="mjd",
                location=EarthLocation.of_site("Cerro Pachon"),
            )
            return get_sun(t).ra.deg

        nside = 32
        scale = 0.9
        order_roll = 1

        # setup a simple test survey with dec < 20
        _, dec = hp.pix2ang(nside, np.arange(hp.nside2npix(nside)), lonlat=True)
        base_footprint = np.zeros(hp.nside2npix(nside), dtype=float)
        wfd_indx = np.where(dec < 20)[0]
        base_footprint[wfd_indx] = 1

        # now test uniformity of uniform rolling and baseline at a range
        # of survey start times
        for mjd_offset in [0, 12, 60, 90, 180, 195, 240, 267, 270, 319]:
            mjd_start = SURVEY_START_MJD + mjd_offset
            sun_ra_start = _get_sun_ra_at_mjd(mjd_start)
            fp_hp = {
                "g": base_footprint.copy(),
                "r": base_footprint.copy(),
                "i": base_footprint.copy(),
            }

            for nslice in [2, 3]:
                uniform_footprint = make_rolling_footprints(
                    fp_hp=fp_hp,
                    mjd_start=mjd_start,
                    sun_ra_start=sun_ra_start,
                    nslice=nslice,
                    scale=scale,
                    nside=nside,
                    wfd_indx=wfd_indx,
                    order_roll=order_roll,
                    n_cycles=None,
                    n_constant_start=2,
                    n_constant_end=6,
                    uniform=True,
                )

                footprint = make_rolling_footprints(
                    fp_hp=fp_hp,
                    mjd_start=mjd_start,
                    sun_ra_start=sun_ra_start,
                    nslice=nslice,
                    scale=scale,
                    nside=nside,
                    wfd_indx=wfd_indx,
                    order_roll=order_roll,
                    n_cycles=None,
                    n_constant_start=3,
                    n_constant_end=6,
                    uniform=False,
                )

                # make sure all of the bands are OK
                for band in ["g", "r", "i"]:
                    # look at each yearly release
                    for i in range(1, 11):
                        mjd = mjd_start + i * 365.25
                        ud = uniform_footprint(mjd, norm=False)
                        uniform_scat = np.std(ud[band][wfd_indx])

                        # uniform rolling surveys have special times where the
                        # scatter in the requested number of observations is 0
                        if nslice == 2:
                            if i in [1, 4, 7, 10]:
                                self.assertTrue(np.allclose(uniform_scat, 0))
                            else:
                                self.assertTrue(uniform_scat != 0)
                        elif nslice == 3:
                            if i in [1, 5, 9, 10]:
                                self.assertTrue(np.allclose(uniform_scat, 0))
                            else:
                                self.assertTrue(uniform_scat != 0)

                        # non-uniform rolling surveys only have uniform
                        # requests at Y1 and Y10
                        d = footprint(mjd, norm=False)
                        scat = np.std(d[band][wfd_indx])

                        if i in [1, 10]:
                            self.assertTrue(np.allclose(scat, 0))
                        else:
                            self.assertTrue(scat != 0)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "DitherDetailer",
    "EuclidDitherDetailer",
    "SplitDetailer",
    "CameraRotDetailer",
    "CameraSmallRotPerObservationListDetailer",
    "ComCamGridDitherDetailer",
    "DeltaCoordDitherDetailer",
)

import sys
import warnings

import numpy as np

from rubin_scheduler.scheduler.detailers import BaseDetailer
from rubin_scheduler.scheduler.features import Conditions, NObsCount
from rubin_scheduler.scheduler.utils import ObservationArray, rotx, thetaphi2xyz, wrap_ra_dec, xyz2thetaphi
from rubin_scheduler.utils import (
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    bearing,
    ddf_locations,
    dest_latlon,
    gnomonic_project_tosky,
    rotation_converter,
)


class DitherDetailer(BaseDetailer):
    """
    make a uniform dither pattern. Offset by a maximum radius in a random
    direction. Mostly intended for DDF pointings, the BaseMarkovDF_survey
    class includes dithering for large areas.

    Parameters
    ----------
    max_dither : `float` (0.7)
        The maximum dither size to use (degrees).
    per_night : `bool` (True)
        If true, use the same dither offset for an entire night. If False,
        generate a dither position per visit.
    n_in_night : `int` (1200)
        The number of unique positions to generate in a night.
        Only used if per_night is False.
    big_int : `int`
        A big integer to use for seeding random number generator
        in case one send in a `night` value that is negative.
    """

    def __init__(
        self, max_dither=0.7, per_night=True, n_in_night=1200, nnights=None, big_int=700000, seed=None
    ):

        if nnights is not None:
            warnings.warn("kwarg nnights deprecated", FutureWarning)
        if seed is not None:
            warnings.warn("kwarg seed deprecated", FutureWarning)

        # Check we can run for 10 years no problem
        test = 36525 < big_int < (sys.maxsize - 36525)
        if not test:
            raise ValueError("Value used for big_int won't work for 10 year survey.")

        self.survey_features = {"n_in_night": NObsCount(per_night=True)}

        self.current_night = -np.inf
        self.max_dither = np.radians(max_dither)
        self.n_in_night = n_in_night
        self.per_night = per_night

        self.offset_ra = None
        self.offset_dec = None

        self.current_night = -np.inf
        self.big_int = big_int

    def _new_ang_rad(self, night):

        # Seed with the current night
        if night < 0:
            seed = self.big_int + night
        else:
            seed = night
        rng = np.random.default_rng(seed)
        if self.per_night:
            n_gen = 1
        else:
            n_gen = self.n_in_night

        self.angles = rng.random(n_gen) * 2 * np.pi
        self.radii = self.max_dither * np.sqrt(rng.random(n_gen))

    def _generate_offsets(self, n_offsets, night):

        if self.per_night:
            if night != self.current_night:
                self._new_ang_rad(night)
                self.current_night = night
                angle = self.angles
                radius = self.radii

                self.offsets_ra = radius * np.cos(angle)
                self.offsets_dec = radius * np.sin(angle)
        else:
            if night != self.current_night:
                self._new_ang_rad(night)
                self.current_night = night
            indx_in_night = self.survey_features["n_in_night"].feature
            angle = self.angles[indx_in_night : indx_in_night + n_offsets]
            radius = self.radii[indx_in_night : indx_in_night + n_offsets]
            self.offsets_ra = radius * np.cos(angle)
            self.offsets_dec = radius * np.sin(angle)

        return self.offsets_ra, self.offsets_dec

    def __call__(self, obs_array, conditions):
        if len(obs_array) == 0:
            return obs_array

        # Generate offsets in RA and Dec
        offsets_ra, offsets_dec = self._generate_offsets(len(obs_array), conditions.night)
        new_ra, new_dec = gnomonic_project_tosky(offsets_ra, offsets_dec, obs_array["RA"], obs_array["dec"])

        obs_array["RA"] = new_ra
        obs_array["dec"] = new_dec

        return obs_array


class DeltaCoordDitherDetailer(BaseDetailer):
    """Dither pattern set by user. Each input observation is
    expanded to have an observation at each dither position.

    Parameters
    ----------
    delta_ra : `np.array`
        Angular distances to move on sky in the RA-direction (degrees).
    delta_dec : `np.array`
        Angular distances to move on sky in the dec-direction (degree).
    delta_rotskypos : `np.array`
        Angular shifts to make in the rotskypos values (degrees). Default None
        applies no rotational shift
    """

    def __init__(self, delta_ra, delta_dec, delta_rotskypos=None):

        if delta_rotskypos is not None:
            if np.size(delta_ra) != np.size(delta_rotskypos):
                raise ValueError(
                    "size of delta_ra (%i) is not equal to size of delta_rotskypos (%i)"
                    % (np.size(delta_ra), np.size(delta_rotskypos))
                )
        if np.size(delta_ra) != np.size(delta_dec):
            raise ValueError(
                "Sizes of delta_ra (%i) and delta_dec (%i) do not match."
                % (np.size(delta_ra), np.size(delta_dec))
            )

        self.survey_features = {}
        self.delta_ra = np.radians(delta_ra)
        self.delta_dec = np.radians(delta_dec)
        if delta_rotskypos is None:
            self.delta_rotskypos = delta_rotskypos
        else:
            self.delta_rotskypos = np.radians(delta_rotskypos)

    def __call__(self, obs_array, conditions):

        output_array_list = []
        for obs in obs_array:
            dithered_observations = ObservationArray(self.delta_ra.size)
            for key in obs.dtype.names:
                dithered_observations[key] = obs[key]

            # Generate grid on the equator (so RA offsets are not small)
            new_decs = self.delta_dec.copy()
            # Adding 90 deg here for later rotation about x-axis
            new_ras = self.delta_ra + np.pi / 2

            # Rotate ra and dec about the x-axis
            # pi/2 term to convert dec to phi
            x, y, z = thetaphi2xyz(new_ras, new_decs + np.pi / 2.0)
            # rotate so offsets are now centered on proper dec
            xp, yp, zp = rotx(obs["dec"], x, y, z)
            theta, phi = xyz2thetaphi(xp, yp, zp)

            new_decs = phi - np.pi / 2
            # Remove 90 degree rot from earlier and center on proper RA
            new_ras = theta - np.pi / 2 + obs["RA"]

            # Make sure coords are in proper range
            new_ras, new_decs = wrap_ra_dec(new_ras, new_decs)

            dithered_observations["RA"] = new_ras
            dithered_observations["dec"] = new_decs
            if self.delta_rotskypos is not None:
                dithered_observations["rotSkyPos_desired"] += self.delta_rotskypos
            output_array_list.append(dithered_observations)

        result = np.concatenate(output_array_list)
        return result


class EuclidDitherDetailer(BaseDetailer):
    """Directional dithering for Euclid DDFs

    Parameters
    ----------
    dither_bearing_dir : `list`
        A list with the dither amplitude in along the axis
        connecting the two positions. Default [-0.25, 1] in degrees.
    dither_bearing_perp : `list`
        A list with the dither amplitude perpendicular to the
        axis connecting the two positions. Default [-0.25, 1]
        in degrees.
    seed : `float`
        Random number seed to use (42).
    per_night : `bool`
        If dither shifts should be per night (default True),
        or if dithers should be every pointing (False).
    ra_a, dec_a, ra_b, dec_b : `float`
        Positions for the two field centers. Default None
        will load the positions from
        rubin_scheduler.utils.ddf_locations
    n_in_night : `int`
        Number of positions to generate dither positions for
        within a night.
    """

    def __init__(
        self,
        dither_bearing_dir=[-0.25, 1],
        dither_bearing_perp=[-0.25, 0.25],
        per_night=True,
        ra_a=None,
        dec_a=None,
        ra_b=None,
        dec_b=None,
        n_in_night=1200,
        nnights=None,
        seed=None,
    ):
        if nnights is not None:
            warnings.warn("kwarg nnights deprecated", FutureWarning)
        if seed is not None:
            warnings.warn("kwarg seed deprecated", FutureWarning)
        self.n_in_night = n_in_night
        self.survey_features = {"n_in_night": NObsCount(per_night=True)}

        default_locations = ddf_locations()

        if ra_a is None:
            self.ra_a = np.radians(default_locations["EDFS_a"][0])
        else:
            self.ra_a = np.radians(ra_a)
        if ra_b is None:
            self.ra_b = np.radians(default_locations["EDFS_b"][0])
        else:
            self.ra_b = np.radians(ra_b)
        if dec_a is None:
            self.dec_a = np.radians(default_locations["EDFS_a"][1])
        else:
            self.dec_a = np.radians(dec_a)
        if dec_b is None:
            self.dec_b = np.radians(default_locations["EDFS_b"][1])
        else:
            self.dec_b = np.radians(dec_b)

        self.dither_bearing_dir = np.radians(dither_bearing_dir)
        self.dither_bearing_perp = np.radians(dither_bearing_perp)

        self.bearing_atob = bearing(self.ra_a, self.dec_a, self.ra_b, self.dec_b)
        self.bearing_btoa = bearing(self.ra_b, self.dec_b, self.ra_a, self.dec_a)

        self.current_night = -np.inf

        self.per_night = per_night
        self.shifted_ra_a = None
        self.shifted_dec_a = None
        self.shifted_ra_b = None
        self.shifted_dec_b = None

    def _new_ang_rad(self, night):

        # seed with the current night
        rng = np.random.default_rng(night)
        if self.per_night:
            n_gen = 1
        else:
            n_gen = self.n_in_night

        self.bearings_mag_1 = rng.uniform(
            low=self.dither_bearing_dir.min(),
            high=self.dither_bearing_dir.max(),
            size=n_gen,
        )
        self.perp_mag_1 = rng.uniform(
            low=self.dither_bearing_perp.min(),
            high=self.dither_bearing_perp.max(),
            size=n_gen,
        )

        self.bearings_mag_2 = rng.uniform(
            low=self.dither_bearing_dir.min(),
            high=self.dither_bearing_dir.max(),
            size=n_gen,
        )
        self.perp_mag_2 = rng.uniform(
            low=self.dither_bearing_perp.min(),
            high=self.dither_bearing_perp.max(),
            size=n_gen,
        )

    def _generate_offsets(self, n_offsets, night):

        indx_in_night = self.survey_features["n_in_night"].feature
        if night != self.current_night:
            self.indx_in_night = 0
            self._new_ang_rad(night)
            self.current_night = night
        if self.per_night:
            bearing_mag = self.bearings_mag_1
            perp_mag = self.perp_mag_1
        else:
            bearing_mag = self.bearings_mag_1[indx_in_night : indx_in_night + n_offsets]
            perp_mag = self.perp_mag_1[indx_in_night : indx_in_night + n_offsets]

        # Move point a along the bearings
        self.shifted_dec_a, self.shifted_ra_a = dest_latlon(
            bearing_mag, self.bearing_atob, self.dec_a, self.ra_a
        )
        self.shifted_dec_a, self.shifted_ra_a = dest_latlon(
            perp_mag,
            self.bearing_atob + np.pi / 2.0,
            self.shifted_dec_a,
            self.shifted_ra_a,
        )

        # Shift the second position
        if self.per_night:
            bearing_mag = self.bearings_mag_2
            perp_mag = self.perp_mag_2
        else:
            bearing_mag = self.bearings_mag_2[indx_in_night : indx_in_night + n_offsets]
            perp_mag = self.perp_mag_2[indx_in_night : indx_in_night + n_offsets]

        self.shifted_dec_b, self.shifted_ra_b = dest_latlon(
            bearing_mag, self.bearing_btoa, self.dec_b, self.ra_b
        )
        self.shifted_dec_b, self.shifted_ra_b = dest_latlon(
            perp_mag,
            self.bearing_btoa + np.pi / 2.0,
            self.shifted_dec_b,
            self.shifted_ra_b,
        )

        return (
            self.shifted_ra_a,
            self.shifted_dec_a,
            self.shifted_ra_b,
            self.shifted_dec_b,
        )

    def __call__(self, obs_array, conditions):
        if np.size(obs_array) == 0:
            return obs_array

        # Generate offsets in RA and Dec
        ra_a, dec_a, ra_b, dec_b = self._generate_offsets(len(obs_array), conditions.night)

        ra = []
        dec = []
        for i, obs in enumerate(obs_array):
            if "DD:EDFS_a" in obs["scheduler_note"][0:9]:
                ra.append(ra_a)
                dec.append(dec_a)
            elif "DD:EDFS_b" in obs["scheduler_note"][0:9]:
                ra.append(ra_b)
                dec.append(dec_b)
            else:
                raise ValueError("scheduler_note does not contain EDFS_a or EDFS_b.")
        obs_array["RA"] = np.concatenate(ra)
        obs_array["dec"] = np.concatenate(dec)
        return obs_array


class SplitDetailer(BaseDetailer):
    """Combine two detailers, but choose which one of them to use, based on
    the presence of a specified string in the scheduler_note.

    Useful to identify different kinds of observations in a ScriptedSurvey
    and then apply different detailers (such as EuclidDetailer vs. standard
    DDF dither detailer).

    Parameters
    ----------
    det1 : `detailers.BaseDetailer`
        The first detailer, to use if the `split_str` is not present
        in scheduler_note.
    det2 : `detailers.BaseDetailer`
        The second detailer, to use if `split_str` is present.
    split_str : `str`
        Search for this string in scheduler_note (matches sub-string).
    """

    def __init__(
        self,
        det1: BaseDetailer,
        det2: BaseDetailer,
        split_str: str = "EDFS",
    ):
        self.det1 = det1
        self.det2 = det2
        self.split_str = split_str

    def __call__(self, observation_array: ObservationArray, conditions: Conditions) -> ObservationArray:
        string_in = [self.split_str in note for note in observation_array["scheduler_note"]]
        string_out = np.logical_not(string_in)

        observation_array[string_out] = self.det1(observation_array[string_out], conditions)
        observation_array[string_in] = self.det2(observation_array[string_in], conditions)
        return observation_array


class CameraRotDetailer(BaseDetailer):
    """
    Randomly set the camera rotation, either for each exposure,
    for each time the detailer is called, or per night.

    Note dithering for each exposure may not be repeatable if
    scheduler is restarted.

    Parameters
    ----------
    max_rot : `float` (90.)
        The maximum amount to offset the camera (degrees)
    min_rot : `float` (90)
        The minimum to offset the camera (degrees)
    dither : `str`
        If "night", change positions per night. If call, change per call.
        If "all", randomize per visit. Default "night".
    telescope : `str`
        Telescope name. Options of "rubin" or "auxtel". Default "rubin".
    n_in_night : `int`
        Number of dither positions to generate. Should be greater than
        the maximum number of visits expected in a night (if
        dither set to "call"), or greather than number of ngihts
        in survey if dither set to "night". Default 7350.
    """

    def __init__(
        self,
        max_rot=90.0,
        min_rot=-90.0,
        dither="night",
        per_night=None,
        seed=42,
        nnights=None,
        telescope="rubin",
        n_in_night=7350,
    ):
        self.survey_features = {}

        if nnights is not None:
            warnings.warn("nnights deprecated", FutureWarning)

        if per_night is True:
            warnings.warn("per_night deprecated, setting dither='night'", FutureWarning)
            dither = "night"
        if per_night is False:
            warnings.warn("per_night deprecated, setting dither='all'", FutureWarning)
            dither = "all"

        if dither is True:
            warnings.warn("dither=True deprecated, swapping to dither='night'", FutureWarning)
            dither = "night"

        if dither is False:
            warnings.warn("dither=False deprecated, swapping to dither='all'", FutureWarning)
            dither = "all"

        self.current_night = -np.inf
        self.max_rot = np.radians(max_rot)
        self.min_rot = np.radians(min_rot)
        self.range = self.max_rot - self.min_rot
        self.dither = dither
        self.rng = np.random.default_rng(seed)
        self.n_in_night = n_in_night
        if dither == "call":
            self.offsets = None
        else:
            self.offsets = self.rng.random(n_in_night)

        self.offset = None
        self.rc = rotation_converter(telescope=telescope)

        self.call_num = 0

    def _generate_offsets(self, n_offsets, night):
        if self.dither == "night":
            if night != self.current_night:
                self.current_night = night
                self.offset = self.offsets[night] * self.range + self.min_rot
            offsets = np.ones(n_offsets) * self.offset
        elif self.dither == "call":
            if night != self.current_night:
                self.current_night = night
                self.call_num = 0
                self.rng = np.random.default_rng(night)
                self.offsets = self.rng.random(self.n_in_night)
            offsets = self.offsets[self.call_num] * self.range + self.min_rot
            self.call_num += 1
            offsets = np.ones(n_offsets) * offsets
        elif self.dither == "all":
            self.rng = np.random.default_rng()
            offsets = self.rng.random(n_offsets) * self.range + self.min_rot
        else:
            raise ValueError("dither kwarg must be set to 'night', 'call', or 'all'.")

        return offsets

    def __call__(self, observation_array, conditions):
        # Generate offsets in camamera rotator
        offsets = self._generate_offsets(len(observation_array), conditions.night)

        alt, az = _approx_ra_dec2_alt_az(
            observation_array["RA"],
            observation_array["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
        )
        obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
        observation_array["rotSkyPos"] = self.rc._rottelpos2rotskypos(offsets, obs_pa)
        observation_array["rotTelPos"] = offsets

        return observation_array


class CameraSmallRotPerObservationListDetailer(BaseDetailer):
    """
    Randomly set the camera rotation for each observation list.

    Generates a small sequential offset for sequential visits
    in the same band; adds a random offset for each band change.

    Parameters
    ----------
    max_rot : `float`, optional
        The maximum amount to offset the camera (degrees).
        Default of 85 allows some padding for camera rotator.
    min_rot : `float`, optional
        The minimum to offset the camera (degrees)
        Default of -85 allows some padding for camera rotator.
    seed : `int`, optional
        Seed for random number generation (per night).
    per_visit_rot : `float`, optional
        Sequential rotation to add per visit.
    telescope : `str`, optional
        Telescope name. Options of "rubin" or "auxtel". Default "rubin".
        This is used to determine conversions between rotSkyPos and rotTelPos.
    """

    def __init__(self, max_rot=85.0, min_rot=-85.0, seed=42, per_visit_rot=0.0, telescope="rubin"):
        self.survey_features = {}

        self.current_night = -1
        self.max_rot = np.radians(max_rot)
        self.min_rot = np.radians(min_rot)
        self.rot_range = self.max_rot - self.min_rot
        self.seed = seed
        self.per_visit_rot = np.radians(per_visit_rot)
        self.offset = None
        self.rc = rotation_converter(telescope=telescope)

    def _generate_offsets_band_change(self, band_list, mjd, initial_offset):
        """Generate a random camera rotation for each band change
        or add a small offset for each sequential observation.
        """
        mjd_hash = round(100 * (np.asarray(mjd).item() % 100))
        rng = np.random.default_rng(mjd_hash * self.seed)

        offsets = np.zeros(len(band_list))

        # Find the locations of the band changes
        band_changes = np.where(np.array(band_list[:-1]) != np.array(band_list[1:]))[0]
        band_changes = np.concatenate([np.array([-1]), band_changes])
        # But add one because of counting and offsets above.
        band_changes += 1
        # Count visits per band in the sequence.
        nvis_per_band = np.concatenate(
            [np.diff(band_changes), np.array([len(band_list) - 1 - band_changes[-1]])]
        )
        # Set up the random rotator offsets for each band change
        # This includes first rotation .. maybe not needed?
        for fchange_idx, nvis_f in zip(band_changes, nvis_per_band):
            rot_range = self.rot_range - self.per_visit_rot * nvis_f
            # At the band change spot, update to random offset
            offsets[fchange_idx:] = rng.random() * rot_range + self.min_rot
            # After the band change point, add incremental rotation
            # (we'll wipe this when we get to next fchange_idx)
            offsets[fchange_idx:] += self.per_visit_rot * np.arange(len(band_list) - fchange_idx)

        offsets = np.where(offsets > self.max_rot, self.max_rot, offsets)
        return offsets

    def __call__(self, observation_list, conditions):
        # Generate offsets in camera rotator
        band_list = [np.asarray(obs["band"]).item() for obs in observation_list]
        offsets = self._generate_offsets_band_change(band_list, conditions.mjd, conditions.rot_tel_pos)

        for i, obs in enumerate(observation_list):
            alt, az = _approx_ra_dec2_alt_az(
                obs["RA"],
                obs["dec"],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
            )
            obs_pa = _approx_altaz2pa(alt, az, conditions.site.latitude_rad)
            obs["rotSkyPos"] = self.rc._rottelpos2rotskypos(offsets[i], obs_pa)
            obs["rotTelPos"] = offsets[i]

        return observation_list


class ComCamGridDitherDetailer(BaseDetailer):
    """
    Generate an offset pattern to synthesize a 2x2 grid of ComCam pointings.

    Parameters
    ----------
    rotTelPosDesired : `float`, (0.)
        The physical rotation angle of the camera rotator (degrees)
    scale : `float` (0.355)
        Half of the offset between grid pointing centers. (degrees)
    dither : `float` (0.05)
        Dither offsets within grid to fill chip gaps. (degrees)
    telescope : `str`, ("comcam")
        Telescope name. Default "comcam".
        This is used to determine conversions between rotSkyPos and rotTelPos.
    """

    def __init__(self, rotTelPosDesired=0.0, scale=0.355, dither=0.05, telescope="comcam"):
        self.survey_features = {}

        self.rotTelPosDesired = np.radians(rotTelPosDesired)
        self.scale = np.radians(scale)
        self.dither = np.radians(dither)
        self.rc = rotation_converter(telescope=telescope)

    def _rotate(self, x, y, angle):
        x_rot = x * np.cos(angle) - y * np.sin(angle)
        y_rot = x * np.sin(angle) + y * np.cos(angle)
        return x_rot, y_rot

    def _generate_offsets(self, n_offsets, band_list, rotSkyPos):
        # 2 x 2 pointing grid
        x_grid = np.array([-1.0 * self.scale, -1.0 * self.scale, self.scale, self.scale])
        y_grid = np.array([-1.0 * self.scale, self.scale, self.scale, -1.0 * self.scale])
        x_grid_rot, y_grid_rot = self._rotate(x_grid, y_grid, -1.0 * rotSkyPos)
        offsets_grid_rot = np.array([x_grid_rot, y_grid_rot]).T

        # Dither pattern within grid to fill chip gaps
        # Psuedo-random offsets
        x_dither = np.array(
            [
                0.0,
                -0.5 * self.dither,
                -1.25 * self.dither,
                1.5 * self.dither,
                0.75 * self.dither,
            ]
        )
        y_dither = np.array(
            [
                0.0,
                -0.75 * self.dither,
                1.5 * self.dither,
                1.25 * self.dither,
                -0.5 * self.dither,
            ]
        )
        x_dither_rot, y_dither_rot = self._rotate(x_dither, y_dither, -1.0 * rotSkyPos)
        offsets_dither_rot = np.array([x_dither_rot, y_dither_rot]).T

        # Find the indices of the band changes
        band_changes = np.where(np.array(band_list[:-1]) != np.array(band_list[1:]))[0]
        band_changes = np.concatenate([np.array([-1]), band_changes])
        band_changes += 1

        offsets = []
        index_band = 0
        for ii in range(0, n_offsets):
            if ii in band_changes:
                # Reset the count after each band change
                index_band = 0

            index_grid = index_band % 4
            index_dither = np.floor(index_band / 4).astype(int) % 5
            offsets.append(offsets_grid_rot[index_grid] + offsets_dither_rot[index_dither])
            index_band += 1

        return np.vstack(offsets)

    def __call__(self, observation_list, conditions):
        if len(observation_list) == 0:
            return observation_list

        band_list = [np.asarray(obs["band"]).item() for obs in observation_list]

        # Initial estimate of rotSkyPos corresponding to desired rotTelPos
        alt, az, pa = _approx_ra_dec2_alt_az(
            observation_list[0]["RA"],
            observation_list[0]["dec"],
            conditions.site.latitude_rad,
            conditions.site.longitude_rad,
            conditions.mjd,
            return_pa=True,
        )
        rotSkyPos = self.rc._rottelpos2rotskypos(self.rotTelPosDesired, pa)

        # Generate offsets in RA and Dec
        offsets = self._generate_offsets(len(observation_list), band_list, rotSkyPos)

        # Project offsets onto sky
        obs_array = np.concatenate(observation_list)
        new_ra, new_dec = gnomonic_project_tosky(
            offsets[:, 0], offsets[:, 1], obs_array["RA"], obs_array["dec"]
        )
        new_ra, new_dec = wrap_ra_dec(new_ra, new_dec)

        # Update observations
        for ii in range(0, len(observation_list)):
            observation_list[ii]["RA"] = new_ra[ii]
            observation_list[ii]["dec"] = new_dec[ii]

            alt, az, pa = _approx_ra_dec2_alt_az(
                new_ra[ii],
                new_dec[ii],
                conditions.site.latitude_rad,
                conditions.site.longitude_rad,
                conditions.mjd,
                return_pa=True,
            )
            observation_list[ii]["rotSkyPos"] = rotSkyPos
            observation_list[ii]["rotTelPos"] = self.rc._rotskypos2rottelpos(rotSkyPos, pa)

        return observation_list



import unittest
import warnings
from inspect import signature

import numpy as np

from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.utils import SURVEY_START_MJD


class TestConditions(unittest.TestCase):
    def test_conditions(self):
        # Generate a Conditions object with things filled out
        mo = ModelObservatory()
        mo_conditions = mo.return_conditions()

        # Fresh empty conditions
        conditions = Conditions()

        # Find the arguments that need to be set
        auxtel_args = signature(conditions.set_auxtel_info)

        to_pass = []
        for key in auxtel_args.parameters.keys():
            if key != "kwargs":
                to_pass.append(getattr(mo_conditions, key))

        conditions.set_auxtel_info(*to_pass)

        for key in auxtel_args.parameters.keys():
            if key != "kwargs":
                if np.size(getattr(conditions, key)) == 1:
                    assert getattr(conditions, key) == getattr(mo_conditions, key)
                else:
                    assert np.array_equal(
                        getattr(conditions, key), getattr(mo_conditions, key), equal_nan=True
                    )

        # Again for the maintel attrs
        conditions = Conditions()
        maintel_args = signature(conditions.set_maintel_info)

        to_pass = []
        for key in maintel_args.parameters.keys():
            if key != "kwargs":
                to_pass.append(getattr(mo_conditions, key))

        conditions.set_maintel_info(*to_pass)

        for key in maintel_args.parameters.keys():
            if key != "kwargs":
                if (np.size(getattr(conditions, key)) == 1) | (isinstance(getattr(conditions, key), list)):
                    assert getattr(conditions, key) == getattr(mo_conditions, key)
                else:
                    assert np.array_equal(
                        getattr(conditions, key), getattr(mo_conditions, key), equal_nan=True
                    )

        # check we can set some arbitrary attributes
        conditions = Conditions()
        conditions.set_attrs(mjd=62511, moon_alt=0.1)

        with self.assertWarns(Warning):
            conditions.set_attrs(not_an_attribute=10)

        # Test that conditions.night is set as expected
        current_mjd = 60914.73382326572
        conditions.mjd = current_mjd
        print(conditions.mjd, conditions.night)
        expected_night = np.floor(current_mjd - 0.5) - np.floor(SURVEY_START_MJD - 0.5)
        self.assertEqual(expected_night, conditions.night)
        # Test that warning is raised if you try to set night directly
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            conditions.night = 10
            # Verify deprecation warning
            assert len(w) >= 1
            assert issubclass(w[-1].category, (DeprecationWarning, FutureWarning))


if __name__ == "__main__":
    unittest.main()



__all__ = ("ModelObservatory",)

import copy
import warnings

import healpy as hp
import numpy as np
from astropy.coordinates import EarthLocation
from astropy.time import Time

import rubin_scheduler.skybrightness_pre as sb
from rubin_scheduler.data import data_versions
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.model_observatory import KinemModel
from rubin_scheduler.scheduler.utils import smallest_signed_angle

# For backwards compatibility
from rubin_scheduler.site_models import (
    Almanac,
    CloudData,
    ConstantCloudData,
    ConstantSeeingData,
    ScheduledDowntimeData,
    SeeingData,
    SeeingModel,
    UnscheduledDowntimeMoreY1Data,
)
from rubin_scheduler.utils import (
    DEFAULT_NSIDE,
    SURVEY_START_MJD,
    Site,
    _angular_separation,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    _hpid2_ra_dec,
    _ra_dec2_hpid,
    calc_lmst,
    calc_season,
    m5_flat_sed,
    rotation_converter,
)


class ModelObservatory:
    """Generate a realistic telemetry stream for the scheduler in simulations,
    including simulating the acquisition of observations.

    Parameters
    ----------
    nside : `int`, optional
        The healpix nside resolution.
        Default None uses `set_default_nside()`.
    mjd : `float`, optional
        The MJD to start the model observatory for observations.
        Used to set current conditions and load sky.
        Default None uses mjd_start.
    mjd_start : `float`, optional
        The MJD of the start of the survey.
        This must be set to start of whole survey, for tracking
        purposes. Should be during the day so night will be
        computed properly.
    alt_min : `float`, optional
        The minimum altitude to compute models at (degrees).
    lax_dome : `bool`, optional
        Passed to observatory model. If true, allows dome creep.
    cloud_limit : `float`, optional.
        The limit to stop taking observations if the cloud model returns
        something equal or higher. Default of 0.3 is validated as a
        "somewhat pessimistic" weather downtime choice.
    sim_to_o : `sim_targetoO` object, optional
        If one would like to inject simulated ToOs into the telemetry
        stream. Default None adds no ToOs.
    park_after : `float`, optional
        Park the telescope after a gap longer than park_after (minutes).
        Default 10 minutes is used to park the telescope during downtime.
    init_load_length : `int`, optional
        The length of pre-scheduled sky brightness values to load
        initially (days). The default is 10 days; shorter values
        can be used for quicker load times.
    kinem_model : `~.scheduler.model_observatory.Kinem_model`, optional
        An instantiated rubin_scheduler Kinem_model object.
        Default of None uses a default Kinem_model.
    seeing_db : `str`, optional
        The filename of the seeing data database, if one would like
        to use an alternate seeing database.
        Default None uses the default seeing database.
    seeing_data : `~.site_models.SeeingData`-like, optional
        If one wants to replace the default seeing_data object.
        Should be an object with a
        __call__ method that takes MJD and returns zenith fwhm_500 in
        arcsec. Set to "ideal" to have constant 0.7" seeing.
    cloud_db : `str`, optional
        The filename of the cloud data database.
        Default of None uses the default database from rubin_sim_data.
    cloud_offset_year : `int`, optional
        The year offset to be passed to CloudData. Default 0.
    cloud_data : `~.site_models.CloudData`-like, optional
        If one wants to replace the default cloud data. Should be an
        object with a __call__ method that takes MJD and returns
        cloudy level. Set to "ideal" for  no clouds.
    downtimes : `np.ndarray`, (N,3) or None, optional
        If one wants to replace the default downtimes. Should be a
        np.array with columns of "start" and "end" with MJD values
        and should include both scheduled and unscheduled downtime.
        Set to "ideal" for no downtime. Default of None will use
        the downtime models from `~.site_models.ScheduledDowntime`
        and `~.site_models.UnscheduledDowntime`.
    no_sky : `bool`, optional
        If True, then don't load any skybrightness files.
        Handy if one wants a well filled out Conditions object,
        but doesn't need the sky since that can be slower to load.
        Default False.
    wind_data : ~.site_models.WindData`-like, optional
        If one wants to replace the default wind_data object. Should
        be an object with a __call__ method that takes the time and
        returns a tuple with the wind speed (m/s) and originating
        direction (radians east of north).
        Default of None uses an idealized WindData object with no wind.
    starting_time_key : `str`, optional
        What key in the almanac to use to determine the start of
        observing on a night. Default "sun_n12_setting", e.g., sun
        at -12 degrees and setting. Other options are
        "sun_n18_setting" and "sunset".
        If surveys are not configured to wait until the sun is lower
        in the sky, observing will start as soon as the time passes
        the time returned by this key, in each night.
    ending_time_key : `str`, optional
        What key in the almanac to use to signify it is time to skip
        to the next night. Default "sun_n12_rising", e.g., sun at
        -12 degrees and rising. Other options are "sun_n18_rising"
        and "sunrise".
    sky_az_limits : `list` [[`float`, `float`]]
        A list of lists giving valid azimuth ranges. e.g.,
        [0, 180] would mean all azimuth values are valid, while
        [[0, 90], [270, 360]] or [270, 90] would mean anywhere in
        the south is invalid.  Degrees.
    sky_alt_limits : `list` [[`float`, `float`]]
        A list of lists giving valid altitude ranges. Degrees.
        For both the alt and az limits, if a particular alt (or az)
        value is included in any limit, it is valid for all of them.
        Altitude limits of [[20, 40], [40, 60]] would allow altitudes
        betwen 20-40 and 40-60, but [[20, 40], [40, 60], [20, 86]]
        will allow altitudes anywhere between 20-86 degrees.
    telescope : `str`
        Telescope name for rotation computations. Default "rubin".
    resolve_rotskypos : `bool`
        If a requested value of rotSkyPos is not accessible, try adding
        90 degrees until it works.

    """

    def __init__(
        self,
        nside=DEFAULT_NSIDE,
        mjd=None,
        mjd_start=SURVEY_START_MJD,
        alt_min=5.0,
        lax_dome=True,
        cloud_limit=0.3,
        sim_to_o=None,
        park_after=10.0,
        init_load_length=10,
        kinem_model=None,
        seeing_db=None,
        seeing_data=None,
        cloud_db=None,
        cloud_offset_year=0,
        cloud_data=None,
        downtimes=None,
        no_sky=False,
        wind_data=None,
        starting_time_key="sun_n12_setting",
        ending_time_key="sun_n12_rising",
        sky_alt_limits=None,
        sky_az_limits=None,
        telescope="rubin",
        cloud_maps=None,
        resolve_rotskypos=True,
    ):
        self.nside = nside
        self.resolve_rotskypos = resolve_rotskypos

        # Set the time now - mjd
        # and the time of the survey start
        self.mjd_start = mjd_start

        if mjd is None:
            mjd = mjd_start

        self.bandlist = ["u", "g", "r", "i", "z", "y"]

        self.cloud_limit = cloud_limit
        self.no_sky = no_sky
        self.alt_min = np.radians(alt_min)
        self.lax_dome = lax_dome
        self.starting_time_key = starting_time_key
        self.ending_time_key = ending_time_key

        self.sim__to_o = sim_to_o

        self.park_after = park_after / 60.0 / 24.0  # To days

        # Rotation converter
        self.rc = rotation_converter(telescope=telescope)

        # Create an astropy location
        self.site = Site("LSST")
        self.location = EarthLocation(
            lat=self.site.latitude, lon=self.site.longitude, height=self.site.height
        )

        # Set up the almanac - use mjd_start to keep "night" count the same.
        self.almanac = Almanac(mjd_start=self.mjd_start)

        # Load up all the models we need
        # Use mjd_start to ensure models always initialize to the same
        # starting point in time.
        mjd_start_time = Time(self.mjd_start, format="mjd")

        # Set up the downtime
        if isinstance(downtimes, str):
            if downtimes == "ideal":
                self.downtimes = np.array(
                    list(zip([], [])),
                    dtype=list(zip(["start", "end"], [float, float])),
                )
            else:
                warnings.warn("Downtimes should be a string equal to " "'ideal', an array or None")
        elif downtimes is None:
            self.down_nights = []
            self.sched_downtime_data = ScheduledDowntimeData(mjd_start_time)
            self.unsched_downtime_data = UnscheduledDowntimeMoreY1Data(mjd_start_time)

            sched_downtimes = self.sched_downtime_data()
            unsched_downtimes = self.unsched_downtime_data()

            down_starts = []
            down_ends = []
            for dt in sched_downtimes:
                down_starts.append(dt["start"].mjd)
                down_ends.append(dt["end"].mjd)
            for dt in unsched_downtimes:
                down_starts.append(dt["start"].mjd)
                down_ends.append(dt["end"].mjd)

            self.downtimes = np.array(
                list(zip(down_starts, down_ends)),
                dtype=list(zip(["start", "end"], [float, float])),
            )
            self.downtimes.sort(order="start")

            # Make sure there aren't any overlapping downtimes
            diff = self.downtimes["start"][1:] - self.downtimes["end"][0:-1]
            while np.min(diff) < 0:
                # Should be able to do this without a loop, but this works
                for i, dt in enumerate(self.downtimes[0:-1]):
                    if self.downtimes["start"][i + 1] < dt["end"]:
                        new_end = np.max([dt["end"], self.downtimes["end"][i + 1]])
                        self.downtimes[i]["end"] = new_end
                        self.downtimes[i + 1]["end"] = new_end

                good = np.where(self.downtimes["end"] - np.roll(self.downtimes["end"], 1) != 0)
                self.downtimes = self.downtimes[good]
                diff = self.downtimes["start"][1:] - self.downtimes["end"][0:-1]
        else:
            self.downtimes = downtimes

        # Set the wind data
        self.wind_data = wind_data

        # Set up the seeing data
        if seeing_data == "ideal":
            self.seeing_data = ConstantSeeingData()
        elif seeing_data is not None:
            self.seeing_data = seeing_data
        else:
            self.seeing_data = SeeingData(mjd_start_time, seeing_db=seeing_db)
        self.seeing_model = SeeingModel()
        self.seeing_indx_dict = {}
        for i, bandname in enumerate(self.seeing_model.band_list):
            self.seeing_indx_dict[bandname] = i

        self.seeing_fwhm_eff = {}
        for key in self.bandlist:
            self.seeing_fwhm_eff[key] = np.zeros(hp.nside2npix(self.nside), dtype=float)

        # Set up the cloud data
        if cloud_data == "ideal":
            self.cloud_data = ConstantCloudData()
        elif cloud_data is not None:
            self.cloud_data = cloud_data
        else:
            self.cloud_data = CloudData(mjd_start_time, cloud_db=cloud_db, offset_year=cloud_offset_year)

        # Set up the skybrightness
        if not self.no_sky:
            self.sky_model = sb.SkyModelPre(init_load_length=init_load_length)
        else:
            self.sky_model = None

        # Set up the kinematic model
        if kinem_model is None:
            self.observatory = KinemModel(mjd0=self.mjd_start)
        else:
            self.observatory = kinem_model

        # Pick up tel alt and az limits from kwargs
        if sky_alt_limits is not None:
            self.sky_alt_limits = np.radians(sky_alt_limits)
        else:
            self.sky_alt_limits = None
        if sky_az_limits is not None:
            self.sky_az_limits = np.radians(sky_az_limits)
        else:
            self.sky_az_limits = None
        # Add the observatory alt/az limits to the user-defined limits
        # But we do have to be careful that we're not overriding more
        # restrictive limits that were already set.
        # So we'll just keep these separate.
        self.tel_alt_limits = copy.deepcopy(
            [
                self.observatory.telalt_minpos_rad,
                self.observatory.telalt_maxpos_rad,
            ]
        )
        self.tel_az_limits = copy.deepcopy(
            [self.observatory.telaz_minpos_rad, self.observatory.telaz_maxpos_rad]
        )
        # Each of these limits will be treated as hard limits that we don't
        # want pointings to stray into, so add a pad around the values
        self.altaz_limit_pad = np.radians(2.0)

        # Let's make sure we're at an openable MJD
        good_mjd = False
        to_set_mjd = mjd
        while not good_mjd:
            good_mjd, to_set_mjd = self.check_mjd(to_set_mjd)
        self.mjd = to_set_mjd

        # Create the map of the season offsets - this map is constant
        ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(self.nside)))
        ra_deg = np.degrees(ra)
        self.season_map = calc_season(ra_deg, [self.mjd_start], self.mjd_start).flatten()
        # Set the sun_ra_start information, for the rolling footprints
        sun_moon_info = self.almanac.get_sun_moon_positions(self.mjd_start)
        self.sun_ra_start = sun_moon_info["sun_RA"] + 0
        # Conditions object to update and return on request
        # (at present, this is not updated -- recreated, below).
        self.conditions = Conditions(
            nside=self.nside,
        )

        self.obs_id_counter = 0

        self.cloud_maps = cloud_maps

    def get_info(self):
        """
        Returns
        -------
        Array with model versions that were instantiated
        """

        # Could add in the data version
        result = []
        versions = data_versions()
        for key in versions:
            result.append([key, versions[key]])

        return result

    def return_conditions(self):
        """
        Returns
        -------
        rubin_scheduler.scheduler.features.conditions object
        """
        self.conditions = Conditions(
            nside=self.nside,
            mjd=self.mjd,
        )

        # Current time as astropy time
        current_time = Time(self.mjd, format="mjd")

        # Clouds. XXX--just the raw value
        self.conditions.bulk_cloud = self.cloud_data(current_time)

        # use conditions object itself to get approx altitude of each
        # healpx
        alts = self.conditions.alt
        azs = self.conditions.az

        good = np.where(alts > self.alt_min)

        # reset the seeing
        for key in self.seeing_fwhm_eff:
            self.seeing_fwhm_eff[key].fill(np.nan)
        # Use the model to get the seeing at this time and airmasses.
        fwhm_500 = self.seeing_data(current_time)
        self.fwhm_500 = fwhm_500
        seeing_dict = self.seeing_model(fwhm_500, self.conditions.airmass[good])
        fwhm_eff = seeing_dict["fwhmEff"]
        for i, key in enumerate(self.seeing_model.band_list):
            self.seeing_fwhm_eff[key][good] = fwhm_eff[i, :]
        self.conditions.fwhm_eff = self.seeing_fwhm_eff

        # sky brightness
        if self.sky_model is not None:
            self.conditions.skybrightness = self.sky_model.return_mags(self.mjd)

        # Model observatory can continue to only refer to band and not filter
        # However, just to ensure set_auxtel_info and set_maintel_info
        # have something to pass in for filters (if those calls being used):
        self.conditions.mounted_filters = self.observatory.mounted_bands
        self.conditions.current_filter = self.observatory.current_band[0]
        # And technically, these next lines are unnecessary
        self.conditions.mounted_bands = self.observatory.mounted_bands
        self.conditions.current_band = self.observatory.current_band[0]

        # Compute the slewtimes
        slewtimes = np.empty(alts.size, dtype=float)
        slewtimes.fill(np.nan)
        # If there has been a gap, park the telescope
        gap = self.mjd - self.observatory.last_mjd
        if gap > self.park_after:
            self.observatory.park()
        slewtimes[good] = self.observatory.slew_times(
            0.0,
            0.0,
            self.mjd,
            alt_rad=alts[good],
            az_rad=azs[good],
            bandname=self.observatory.current_band,
            lax_dome=self.lax_dome,
            update_tracking=False,
        )
        self.conditions.slewtime = slewtimes

        # Let's get the sun and moon
        sun_moon_info = self.almanac.get_sun_moon_positions(self.mjd)
        # convert these to scalars
        for key in sun_moon_info:
            sun_moon_info[key] = sun_moon_info[key].max()
        self.conditions.moon_phase = sun_moon_info["moon_phase"]

        self.conditions.moon_alt = sun_moon_info["moon_alt"]
        self.conditions.moon_az = sun_moon_info["moon_az"]
        self.conditions.moon_ra = sun_moon_info["moon_RA"]
        self.conditions.moon_dec = sun_moon_info["moon_dec"]
        self.conditions.sun_alt = sun_moon_info["sun_alt"]
        self.conditions.sun_az = sun_moon_info["sun_az"]
        self.conditions.sun_ra = sun_moon_info["sun_RA"]
        self.conditions.sun_dec = sun_moon_info["sun_dec"]

        self.conditions.lmst = calc_lmst(self.mjd, self.site.longitude_rad)

        self.conditions.tel_ra = self.observatory.current_ra_rad
        self.conditions.tel_dec = self.observatory.current_dec_rad
        self.conditions.tel_alt = self.observatory.last_alt_rad
        self.conditions.tel_az = self.observatory.last_az_rad

        self.conditions.rot_tel_pos = self.observatory.last_rot_tel_pos_rad
        self.conditions.cumulative_azimuth_rad = self.observatory.cumulative_azimuth_rad

        # Add in the almanac information
        self.conditions.sunset = self.almanac.sunsets["sunset"][self.almanac_indx]
        self.conditions.sun_n12_setting = self.almanac.sunsets["sun_n12_setting"][self.almanac_indx]
        self.conditions.sun_n18_setting = self.almanac.sunsets["sun_n18_setting"][self.almanac_indx]
        self.conditions.sun_n18_rising = self.almanac.sunsets["sun_n18_rising"][self.almanac_indx]
        self.conditions.sun_n12_rising = self.almanac.sunsets["sun_n12_rising"][self.almanac_indx]
        self.conditions.sunrise = self.almanac.sunsets["sunrise"][self.almanac_indx]
        self.conditions.moonrise = self.almanac.sunsets["moonrise"][self.almanac_indx]
        self.conditions.moonset = self.almanac.sunsets["moonset"][self.almanac_indx]
        sun_moon_info_start_of_night = self.almanac.get_sun_moon_positions(self.conditions.sunset)
        self.conditions.moon_phase_sunset = sun_moon_info_start_of_night["moon_phase"]

        # Telescope limits
        self.conditions.sky_az_limits = self.sky_az_limits
        self.conditions.sky_alt_limits = self.sky_alt_limits
        self.conditions.tel_alt_limits = self.tel_alt_limits
        self.conditions.tel_az_limits = self.tel_az_limits

        # Planet positions from almanac
        self.conditions.planet_positions = self.almanac.get_planet_positions(self.mjd)

        # See if there are any ToOs to include
        if self.sim__to_o is not None:
            toos = self.sim__to_o(self.mjd)
            if toos is not None:
                self.conditions.targets_of_opportunity = toos

        if self.wind_data is not None:
            wind_speed, wind_direction = self.wind_data(current_time)
            self.conditions.wind_speed = wind_speed
            self.conditions.wind_direction = wind_direction

        if self.cloud_maps is not None:
            self.conditions.cloud_maps = self.cloud_maps

        return self.conditions

    @property
    def mjd(self):
        return self._mjd

    @mjd.setter
    def mjd(self, value):
        self._mjd = value
        self.almanac_indx = self.almanac.mjd_indx(value)
        self.night = np.floor(self.mjd - self.mjd_start).astype(int)

    def observation_add_data(self, observation):
        """
        Fill in the metadata for a completed observation
        """
        current_time = Time(self.mjd, format="mjd")

        observation["clouds"] = self.cloud_data(current_time)
        observation["airmass"] = 1.0 / np.cos(np.pi / 2.0 - observation["alt"])
        # Seeing
        fwhm_500 = self.seeing_data(current_time)
        seeing_dict = self.seeing_model(fwhm_500, observation["airmass"])
        observation["FWHMeff"] = seeing_dict["fwhmEff"][self.seeing_indx_dict[observation["band"][0]]]
        observation["FWHM_geometric"] = seeing_dict["fwhmGeom"][self.seeing_indx_dict[observation["band"][0]]]
        observation["FWHM_500"] = fwhm_500

        observation["night"] = self.night
        observation["mjd"] = self.mjd

        if self.sky_model is not None:
            hpid = _ra_dec2_hpid(self.sky_model.nside, observation["RA"], observation["dec"])
            observation["skybrightness"] = self.sky_model.return_mags(
                self.mjd, indx=[hpid], extrapolate=True
            )[observation["band"][0]]

        observation["fivesigmadepth"] = m5_flat_sed(
            observation["band"][0],
            observation["skybrightness"],
            observation["FWHMeff"],
            observation["exptime"] / observation["nexp"],
            observation["airmass"],
            nexp=observation["nexp"],
        )
        # If there is cloud extinction, apply it.
        if self.cloud_maps is not None:
            hpid = _ra_dec2_hpid(self.sky_model.nside, observation["RA"], observation["dec"])
            cloud_extinction = self.cloud_maps.extinction_closest(self.mjd, hpid)
            observation["fivesigmadepth"] -= cloud_extinction
            observation["cloud_extinction"] = cloud_extinction

        lmst = calc_lmst(self.mjd, self.site.longitude_rad)
        observation["lmst"] = lmst

        sun_moon_info = self.almanac.get_sun_moon_positions(self.mjd)
        observation["sunAlt"] = sun_moon_info["sun_alt"]
        observation["sunAz"] = sun_moon_info["sun_az"]
        observation["sunRA"] = sun_moon_info["sun_RA"]
        observation["sunDec"] = sun_moon_info["sun_dec"]
        observation["moonAlt"] = sun_moon_info["moon_alt"]
        observation["moonAz"] = sun_moon_info["moon_az"]
        observation["moonRA"] = sun_moon_info["moon_RA"]
        observation["moonDec"] = sun_moon_info["moon_dec"]
        observation["moonDist"] = _angular_separation(
            observation["RA"],
            observation["dec"],
            observation["moonRA"],
            observation["moonDec"],
        )
        observation["solarElong"] = _angular_separation(
            observation["RA"],
            observation["dec"],
            observation["sunRA"],
            observation["sunDec"],
        )
        observation["moonPhase"] = sun_moon_info["moon_phase"]

        observation["ID"] = self.obs_id_counter
        self.obs_id_counter += 1

        return observation

    def check_up(self, mjd):
        """See if we are in downtime

        Returns
        --------
        is_up, mjd : `bool`, `float`
            Returns (True, current_mjd) if telescope is up
            and (False, downtime_ends_mjd) if in downtime
        """

        result = True, mjd
        indx = np.searchsorted(self.downtimes["start"], mjd, side="right") - 1
        if indx >= 0:
            if mjd < self.downtimes["end"][indx]:
                result = False, self.downtimes["end"][indx]
        return result

    def check_mjd(self, mjd, cloud_skip=20.0):
        """See if an mjd is ok to observe

        Parameters
        ----------
        cloud_skip : float (20)
            How much time to skip ahead if it's cloudy (minutes)

        Returns
        -------
        mjd_ok : `bool`
        mdj : `float`
            If True, the input mjd. If false, a good mjd to skip
            forward to.
        """
        passed = True
        new_mjd = mjd + 0

        # Maybe set this to a while loop to make sure we don't
        # land on another cloudy time?
        # or just make this an entire recursive call?
        clouds = self.cloud_data(Time(mjd, format="mjd"))

        if clouds > self.cloud_limit:
            passed = False
            while clouds > self.cloud_limit:
                new_mjd = new_mjd + cloud_skip / 60.0 / 24.0
                clouds = self.cloud_data(Time(new_mjd, format="mjd"))
        alm_indx = np.searchsorted(self.almanac.sunsets[self.starting_time_key], mjd, side="right") - 1
        # at the end of the night, advance to the next setting twilight
        if mjd > self.almanac.sunsets[self.ending_time_key][alm_indx]:
            passed = False
            new_mjd = self.almanac.sunsets[self.starting_time_key][alm_indx + 1]
        if mjd < self.almanac.sunsets[self.starting_time_key][alm_indx]:
            passed = False
            new_mjd = self.almanac.sunsets[self.starting_time_key][alm_indx + 1]
        # We're in a down time, if down, advance to the end of the downtime
        if not self.check_up(mjd)[0]:
            passed = False
            new_mjd = self.check_up(mjd)[1]
        # recursive call to make sure we skip far enough ahead
        if not passed:
            while not passed:
                passed, new_mjd = self.check_mjd(new_mjd)
            return False, new_mjd
        else:
            return True, mjd

    def _update_rot_sky_pos(self, observation, try_shifts=np.array([-np.pi / 2, np.pi / 2, -np.pi, np.pi])):
        """Update the rotSkyPos value to make sure it is observable.

        Parameters
        ----------
        observation : `rubin_scheduler.scheduler.ObservationArray`
            The observation to fix rotSkyPos on.
        try_shifts : `np.array`
            If the rotSkyPos value results in an invalid rotTelPos,
            rotate by the values in try_shifts and select the
            one that results in the smallest rotator slew.
            Default [-np.pi/2, np.pi/2, -np.pi, np.pi] (radians).
        """

        # Grab the rotator limit from the observatory model
        # Set so default rot_limit = [-90, 90]
        rot_limit = [
            self.observatory.telrot_minpos_rad,
            self.observatory.telrot_maxpos_rad,
        ]

        alt, az = _approx_ra_dec2_alt_az(
            observation["RA"],
            observation["dec"],
            self.site.latitude_rad,
            self.site.longitude_rad,
            self.mjd,
        )

        obs_pa = _approx_altaz2pa(alt, az, self.site.latitude_rad)

        if self.resolve_rotskypos:
            if not np.isfinite(observation["rotSkyPos"]):
                warnings.warn("No finite rotSkyPos value, using rotSkyPos_desired")
                observation["rotSkyPos"] = observation["rotSkyPos_desired"]
            rottelpos = self.rc._rotskypos2rottelpos(observation["rotSkyPos"], obs_pa)

            if (rottelpos < rot_limit[0]) | (rottelpos > rot_limit[1]):
                potential_rottelpos = rottelpos + try_shifts
                valid = np.where((potential_rottelpos > rot_limit[0]) & (potential_rottelpos < rot_limit[1]))[
                    0
                ]
                potential_rottelpos = potential_rottelpos[valid]
                # Which potential rot_tel_pos is closest to the
                # current rot_tel_pos?
                if self.observatory.current_rot_sky_pos_rad is None:
                    current_rot_tel_pos = 0
                else:
                    current_rot_tel_pos = self.rc._rotskypos2rottelpos(
                        self.observatory.current_rot_sky_pos_rad, obs_pa
                    )
                ang_diff = np.abs(smallest_signed_angle(current_rot_tel_pos, potential_rottelpos))
                indx = np.where(ang_diff == np.min(ang_diff))[0]
                rottelpos = potential_rottelpos[indx]

            observation["rotSkyPos"] = self.rc._rottelpos2rotskypos(rottelpos, obs_pa)

        else:
            # If the observation has a rotTelPos set,
            # use it to compute rotSkyPos
            if np.isfinite(observation["rotTelPos"]):
                observation["rotSkyPos"] = self.rc._rottelpos2rotskypos(observation["rotTelPos"], obs_pa)
                observation["rotTelPos"] = np.nan
            else:
                # Try to fall back to rotSkyPos_desired
                possible_rot_tel_pos = self.rc._rotskypos2rottelpos(observation["rotSkyPos_desired"], obs_pa)
                # If in range, use rotSkyPos_desired for rotSkyPos
                if (possible_rot_tel_pos > rot_limit[0]) & (possible_rot_tel_pos < rot_limit[1]):
                    observation["rotSkyPos"] = observation["rotSkyPos_desired"]
                    observation["rotTelPos"] = np.nan
                else:
                    # Fall back to the backup rotation angle if needed.
                    observation["rotSkyPos"] = np.nan
                    observation["rotTelPos"] = observation["rotTelPos_backup"]

        return observation

    def observe(self, observation):
        """Try to make an observation

        Returns
        -------
        observation : observation object
            None if there was no observation taken. Completed
            observation with meta data filled in.
        new_night : bool
            Have we started a new night.
        """

        start_night = self.night.copy()

        observation = self._update_rot_sky_pos(observation)

        # If there has been a long gap, assume telescope stopped
        # tracking and parked
        gap = self.mjd - self.observatory.last_mjd
        if gap > self.park_after:
            self.observatory.park()

        # Compute what alt,az we have tracked to (or are parked at)
        start_alt, start_az, start_rot_tel_pos = self.observatory.current_alt_az(self.mjd)
        # Slew to new position and execute observation. Use the
        # requested rotTelPos position, obsevation['rotSkyPos'] will
        # be ignored.
        slewtime, visittime = self.observatory.observe(
            observation,
            self.mjd,
            rot_tel_pos=observation["rotTelPos"],
            lax_dome=self.lax_dome,
        )

        # inf slewtime means the observation failed (probably outside
        # alt limits)
        if ~np.all(np.isfinite(slewtime)):
            return None, False

        observation_worked, new_mjd = self.check_mjd(self.mjd + (slewtime + visittime) / 24.0 / 3600.0)

        if observation_worked:
            observation["visittime"] = visittime
            observation["slewtime"] = slewtime
            observation["slewdist"] = _angular_separation(
                start_az,
                start_alt,
                self.observatory.last_az_rad,
                self.observatory.last_alt_rad,
            )
            self.mjd = self.mjd + slewtime / 24.0 / 3600.0
            # Reach into the observatory model to pull out the
            # relevant data it has calculated
            # Not bothering to fetch alt,az,pa,rottelpos as those
            # were computed before the slew was executed
            # so will be off by seconds to minutes. And they
            # shouldn't be needed by the scheduler.

            observation["rotSkyPos"] = self.observatory.current_rot_sky_pos_rad
            observation["cummTelAz"] = self.observatory.cumulative_azimuth_rad

            # But we do need the altitude to
            # get the airmass and then 5-sigma depth
            # this altitude should get clobbered later
            # by sim_runner.
            observation["alt"] = self.observatory.last_alt_rad

            # Metadata on observation is after slew and settle,
            # so at start of exposure.
            result = self.observation_add_data(observation)
            self.mjd = self.mjd + visittime / 24.0 / 3600.0
            new_night = False
        else:
            result = None
            self.observatory.park()
            # Skip to next legitimate mjd
            self.mjd = new_mjd
            now_night = self.night
            if now_night == start_night:
                new_night = False
            else:
                new_night = True

        return result, new_night

    # methods to reach through and adjust the kinematic model if desired
    def setup_camera(self, **kwargs):
        self.observatory.setup_camera(**kwargs)

    def setup_dome(self, **kwargs):
        self.observatory.setup_dome(**kwargs)

    def setup_telescope(self, **kwargs):
        self.observatory.setup_telescope(**kwargs)

    def setup_setup_optics(self, **kwargs):
        self.observatory.setup_optics(**kwargs)
